from __future__ import annotations

import asyncio
import enum
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    Optional,
    Sequence,
    TYPE_CHECKING,
)
from uuid import UUID, uuid4

import aiohttp
import graphene
from graphene.types.datetime import DateTime as GQLDateTime
from passlib.hash import bcrypt
import sqlalchemy as sa
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import (
    AsyncConnection as SAConnection,
    AsyncEngine as SAEngine,
)
from sqlalchemy.types import TypeDecorator, VARCHAR

from ai.backend.common.logging import BraceStyleAdapter

from ..api.exceptions import VFolderOperationFailed
from .base import (
    EnumValueType,
    IDColumn,
    Item,
    PaginatedList,
    metadata,
    set_if_set,
    batch_result,
    batch_multiresult,
)
from .storage import StorageSessionManager

if TYPE_CHECKING:
    from .gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger(__file__))


__all__: Sequence[str] = (
    'users',
    'User', 'UserList',
    'UserGroup', 'UserRole',
    'UserInput', 'ModifyUserInput',
    'CreateUser', 'ModifyUser', 'DeleteUser',
    'UserStatus', 'ACTIVE_USER_STATUSES', 'INACTIVE_USER_STATUSES',
)


class PasswordColumn(TypeDecorator):
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        return _hash_password(value)


class UserRole(str, enum.Enum):
    """
    User's role.
    """
    SUPERADMIN = 'superadmin'
    ADMIN = 'admin'
    USER = 'user'
    MONITOR = 'monitor'


class UserStatus(str, enum.Enum):
    """
    User account status.
    """
    ACTIVE = 'active'
    INACTIVE = 'inactive'
    DELETED = 'deleted'
    BEFORE_VERIFICATION = 'before-verification'


ACTIVE_USER_STATUSES = (
    UserStatus.ACTIVE,
)

INACTIVE_USER_STATUSES = (
    UserStatus.INACTIVE,
    UserStatus.DELETED,
    UserStatus.BEFORE_VERIFICATION,
)


users = sa.Table(
    'users', metadata,
    IDColumn('uuid'),
    sa.Column('username', sa.String(length=64), unique=True),
    sa.Column('email', sa.String(length=64), index=True,
              nullable=False, unique=True),
    sa.Column('password', PasswordColumn()),
    sa.Column('need_password_change', sa.Boolean),
    sa.Column('full_name', sa.String(length=64)),
    sa.Column('description', sa.String(length=500)),
    sa.Column('status', EnumValueType(UserStatus), default=UserStatus.ACTIVE, nullable=False),
    sa.Column('status_info', sa.Unicode(), nullable=True, default=sa.null()),
    sa.Column('created_at', sa.DateTime(timezone=True),
              server_default=sa.func.now()),
    sa.Column('modified_at', sa.DateTime(timezone=True),
              server_default=sa.func.now(), onupdate=sa.func.current_timestamp()),
    #: Field for synchronization with external services.
    sa.Column('integration_id', sa.String(length=512)),

    sa.Column('domain_name', sa.String(length=64),
              sa.ForeignKey('domains.name'), index=True),
    sa.Column('role', EnumValueType(UserRole), default=UserRole.USER),
)


class UserGroup(graphene.ObjectType):
    id = graphene.UUID()
    name = graphene.String()

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: Row) -> Optional[UserGroup]:
        if row is None:
            return None
        return cls(
            id=row['id'],
            name=row['name'],
        )

    @classmethod
    async def batch_load_by_user_id(cls, ctx: GraphQueryContext, user_ids: Sequence[UUID]):
        async with ctx.db.begin() as conn:
            from .group import groups, association_groups_users as agus
            j = agus.join(groups, agus.c.group_id == groups.c.id)
            query = (
                sa.select([agus.c.user_id, groups.c.name, groups.c.id])
                .select_from(j)
                .where(agus.c.user_id.in_(user_ids))
            )
            return await batch_multiresult(
                ctx, conn, query, cls,
                user_ids, lambda row: row['user_id'],
            )


class User(graphene.ObjectType):

    class Meta:
        interfaces = (Item, )

    uuid = graphene.UUID()  # legacy
    username = graphene.String()
    email = graphene.String()
    password = graphene.String()
    need_password_change = graphene.Boolean()
    full_name = graphene.String()
    description = graphene.String()
    is_active = graphene.Boolean()
    status = graphene.String()
    status_info = graphene.String()
    created_at = GQLDateTime()
    modified_at = GQLDateTime()
    domain_name = graphene.String()
    role = graphene.String()

    groups = graphene.List(lambda: UserGroup)

    async def resolve_groups(
        self,
        info: graphene.ResolveInfo,
    ) -> Iterable[UserGroup]:
        ctx: GraphQueryContext = info.context
        manager = ctx.dataloader_manager
        loader = manager.get_loader(ctx, 'UserGroup.by_user_id')
        return await loader.load(self.id)

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> User:
        return cls(
            id=row['uuid'],
            uuid=row['uuid'],
            username=row['username'],
            email=row['email'],
            need_password_change=row['need_password_change'],
            full_name=row['full_name'],
            description=row['description'],
            is_active=True if row['status'] == UserStatus.ACTIVE else False,  # legacy
            status=row['status'],
            status_info=row['status_info'],
            created_at=row['created_at'],
            modified_at=row['modified_at'],
            domain_name=row['domain_name'],
            role=row['role'],
        )

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        domain_name: str = None,
        group_id: UUID = None,
        is_active: bool = None,
        status: str = None,
        limit: int = None,
    ) -> Sequence[User]:
        """
        Load user's information. Group names associated with the user are also returned.
        """
        if group_id is not None:
            from .group import association_groups_users as agus
            j = (users.join(agus, agus.c.user_id == users.c.uuid))
            query = (
                sa.select([users])
                .select_from(j)
                .where(agus.c.group_id == group_id)
            )
        else:
            query = (
                sa.select([users])
                .select_from(users)
            )
        if ctx.user['role'] != UserRole.SUPERADMIN:
            query = query.where(users.c.domain_name == ctx.user['domain_name'])
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if status is not None:
            query = query.where(users.c.status == UserStatus(status))
        elif is_active is not None:  # consider is_active field only if status is empty
            _statuses = ACTIVE_USER_STATUSES if is_active else INACTIVE_USER_STATUSES
            query = query.where(users.c.status.in_(_statuses))
        if limit is not None:
            query = query.limit(limit)
        return [cls.from_row(ctx, row) async for row in (await ctx.db_conn.stream(query))]

    @staticmethod
    async def load_count(
        ctx: GraphQueryContext,
        *,
        domain_name: str = None,
        group_id: UUID = None,
        is_active: bool = None,
        status: str = None,
    ) -> int:
        if group_id is not None:
            from .group import association_groups_users as agus
            j = (users.join(agus, agus.c.user_id == users.c.uuid))
            query = (
                sa.select([sa.func.count(users.c.uuid)])
                .select_from(j)
                .where(agus.c.group_id == group_id)
            )
        else:
            query = (
                sa.select([sa.func.count(users.c.uuid)])
                .select_from(users)
            )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if status is not None:
            query = query.where(users.c.status == UserStatus(status))
        elif is_active is not None:  # consider is_active field only if status is empty
            _statuses = ACTIVE_USER_STATUSES if is_active else INACTIVE_USER_STATUSES
            query = query.where(users.c.status.in_(_statuses))
        result = await ctx.db_conn.execute(query)
        return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: str = None,
        group_id: UUID = None,
        is_active: bool = None,
        status: str = None,
        order_key: str = None,
        order_asc: bool = True,
    ) -> Sequence[User]:
        if order_key is None:
            _ordering = sa.desc(users.c.created_at)
        else:
            _order_func = sa.asc if order_asc else sa.desc
            _ordering = _order_func(getattr(users.c, order_key))
        if group_id is not None:
            from .group import association_groups_users as agus
            j = (users.join(agus, agus.c.user_id == users.c.uuid))
            query = (
                sa.select([users])
                .select_from(j)
                .where(agus.c.group_id == group_id)
                .order_by(_ordering)
                .limit(limit)
                .offset(offset)
            )
        else:
            query = (
                sa.select([users])
                .select_from(users)
                .order_by(_ordering)
                .limit(limit)
                .offset(offset)
            )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if status is not None:
            query = query.where(users.c.status == UserStatus(status))
        elif is_active is not None:  # consider is_active field only if status is empty
            _statuses = ACTIVE_USER_STATUSES if is_active else INACTIVE_USER_STATUSES
            query = query.where(users.c.status.in_(_statuses))
        return [
            cls.from_row(ctx, row) async for row in (await ctx.db_conn.stream(query))
        ]

    @classmethod
    async def batch_load_by_email(
        cls,
        ctx: GraphQueryContext,
        emails: Sequence[str] = None,
        *,
        domain_name: str = None,
        is_active: bool = None,
        status: str = None,
    ) -> Sequence[Optional[User]]:
        if not emails:
            return []
        query = (
            sa.select([users])
            .select_from(users)
            .where(users.c.email.in_(emails))
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if status is not None:
            query = query.where(users.c.status == UserStatus(status))
        elif is_active is not None:  # consider is_active field only if status is empty
            _statuses = ACTIVE_USER_STATUSES if is_active else INACTIVE_USER_STATUSES
            query = query.where(users.c.status.in_(_statuses))
        return await batch_result(
            ctx, ctx.db_conn, query, cls,
            emails, lambda row: row['email'],
        )

    @classmethod
    async def batch_load_by_uuid(
        cls,
        ctx: GraphQueryContext,
        user_ids: Sequence[UUID] = None,
        *,
        domain_name: str = None,
        is_active: bool = None,
        status: str = None,
    ) -> Sequence[Optional[User]]:
        if not user_ids:
            return []
        query = (
            sa.select([users])
            .select_from(users)
            .where(users.c.uuid.in_(user_ids))
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if status is not None:
            query = query.where(users.c.status == UserStatus(status))
        elif is_active is not None:  # consider is_active field only if status is empty
            _statuses = ACTIVE_USER_STATUSES if is_active else INACTIVE_USER_STATUSES
            query = query.where(users.c.status.in_(_statuses))
        return await batch_result(
            ctx, ctx.db_conn, query, cls,
            user_ids, lambda row: row['uuid'],
        )


class UserList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(User, required=True)


class UserInput(graphene.InputObjectType):
    username = graphene.String(required=True)
    password = graphene.String(required=True)
    need_password_change = graphene.Boolean(required=True)
    full_name = graphene.String(required=False, default='')
    description = graphene.String(required=False, default='')
    is_active = graphene.Boolean(required=False, default=True)
    status = graphene.String(required=False, default=UserStatus.ACTIVE)
    domain_name = graphene.String(required=True, default='default')
    role = graphene.String(required=False, default=UserRole.USER)
    group_ids = graphene.List(lambda: graphene.String, required=False)

    # When creating, you MUST set all fields.
    # When modifying, set the field to "None" to skip setting the value.


class ModifyUserInput(graphene.InputObjectType):
    username = graphene.String(required=False)
    password = graphene.String(required=False)
    need_password_change = graphene.Boolean(required=False)
    full_name = graphene.String(required=False)
    description = graphene.String(required=False)
    is_active = graphene.Boolean(required=False)
    status = graphene.String(required=False)
    domain_name = graphene.String(required=False)
    role = graphene.String(required=False)
    group_ids = graphene.List(lambda: graphene.String, required=False)


class PurgeUserInput(graphene.InputObjectType):
    purge_shared_vfolders = graphene.Boolean(required=False, default=False)


class CreateUser(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        email = graphene.String(required=True)
        props = UserInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    user = graphene.Field(lambda: User, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        email: str,
        props: UserInput,
    ) -> CreateUser:
        graph_ctx: GraphQueryContext = info.context
        username = props.username if props.username else email
        if props.status is None and props.is_active is not None:
            _status = UserStatus.ACTIVE if props.is_active else UserStatus.INACTIVE
        else:
            _status = UserStatus(props.status)
        data = {
            'username': username,
            'email': email,
            'password': props.password,
            'need_password_change': props.need_password_change,
            'full_name': props.full_name,
            'description': props.description,
            'status': _status,
            'status_info': 'admin-requested',  # user mutation is only for admin
            'domain_name': props.domain_name,
            'role': UserRole(props.role),
        }
        try:
            query = (users.insert().values(data))
            result = await graph_ctx.db_conn.execute(query)
            if result.rowcount > 0:
                # Read the created user data from DB.
                checkq = users.select().where(users.c.email == email)
                result = await graph_ctx.db_conn.execute(checkq)
                o = User.from_row(info.context, result.first())

                # Create user's first access_key and secret_key.
                from .keypair import generate_keypair, generate_ssh_keypair, keypairs
                ak, sk = generate_keypair()
                pubkey, privkey = generate_ssh_keypair()
                is_admin = True if data['role'] in [UserRole.SUPERADMIN, UserRole.ADMIN] else False
                kp_data = {
                    'user_id': email,
                    'access_key': ak,
                    'secret_key': sk,
                    'is_active': True if _status == UserStatus.ACTIVE else False,
                    'is_admin': is_admin,
                    'resource_policy': 'default',
                    'concurrency_used': 0,
                    'rate_limit': 10000,
                    'num_queries': 0,
                    'user': o.uuid,
                    'ssh_public_key': pubkey,
                    'ssh_private_key': privkey,
                }
                query = (keypairs.insert().values(kp_data))
                await graph_ctx.db_conn.execute(query)

                # Add user to groups if group_ids parameter is provided.
                from .group import association_groups_users, groups
                if props.group_ids:
                    query = (sa.select([groups.c.id])
                                .select_from(groups)
                                .where(groups.c.domain_name == props.domain_name)
                                .where(groups.c.id.in_(props.group_ids)))
                    result = await graph_ctx.db_conn.execute(query)
                    grps = result.fetchall()
                    if grps:
                        values = [{'user_id': o.uuid, 'group_id': grp.id} for grp in grps]
                        query = association_groups_users.insert().values(values)
                        await graph_ctx.db_conn.execute(query)
                return cls(ok=True, msg='success', user=o)
            else:
                return cls(ok=False, msg='failed to create user', user=None)
        except sa.exc.IntegrityError as e:
            return cls(ok=False, msg=f'integrity error: {e}', user=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as e:
            return cls(ok=False, msg=f'unexpected error: {e}', user=None)


class ModifyUser(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        email = graphene.String(required=True)
        props = ModifyUserInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    user = graphene.Field(lambda: User)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        email: str,
        props: ModifyUserInput,
    ) -> ModifyUser:
        graph_ctx: GraphQueryContext = info.context
        data: Dict[str, Any] = {}
        set_if_set(props, data, 'username')
        set_if_set(props, data, 'password')
        set_if_set(props, data, 'need_password_change')
        set_if_set(props, data, 'full_name')
        set_if_set(props, data, 'description')
        set_if_set(props, data, 'status')
        set_if_set(props, data, 'domain_name')
        set_if_set(props, data, 'role')

        if 'role' in data:
            data['role'] = UserRole(data['role'])

        if data.get('status') is None and props.is_active is not None:
            _status = 'active' if props.is_active else 'inactive'
            data['status'] = _status
        if 'status' in data and data['status'] is not None:
            data['status'] = UserStatus(data['status'])

        if not data and not props.group_ids:
            return cls(ok=False, msg='nothing to update', user=None)

        try:
            # Get previous domain name of the user.
            query = (sa.select([users.c.domain_name, users.c.role, users.c.status])
                        .select_from(users)
                        .where(users.c.email == email))
            result = await graph_ctx.db_conn.execute(query)
            row = result.first()
            prev_domain_name = row.domain_name
            prev_role = row.role

            if 'status' in data and row.status != data['status']:
                data['status_info'] = 'admin-requested'  # user mutation is only for admin

            # Update user.
            query = (users.update().values(data).where(users.c.email == email))
            result = await graph_ctx.db_conn.execute(query)
            if result.rowcount > 0:
                checkq = users.select().where(users.c.email == email)
                result = await graph_ctx.db_conn.execute(checkq)
                o = User.from_row(graph_ctx, result.first())
            else:
                return cls(ok=False, msg='no such user', user=None)

            # Update keypair if user's role is updated.
            # NOTE: This assumes that user have only one keypair.
            if 'role' in data and data['role'] != prev_role:
                from ai.backend.manager.models import keypairs
                query = (sa.select([keypairs.c.user,
                                    keypairs.c.is_active,
                                    keypairs.c.is_admin])
                            .select_from(keypairs)
                            .where(keypairs.c.user == o.uuid)
                            .order_by(sa.desc(keypairs.c.is_admin))
                            .order_by(sa.desc(keypairs.c.is_active)))
                result = await graph_ctx.db_conn.execute(query)
                if data['role'] in [UserRole.SUPERADMIN, UserRole.ADMIN]:
                    # User's becomes admin. Set the keypair as active admin.
                    kp = result.first()
                    kp_data = dict()
                    if not kp.is_admin:
                        kp_data['is_admin'] = True
                    if not kp.is_active:
                        kp_data['is_active'] = True
                    if len(kp_data.keys()) > 0:
                        query = (keypairs.update()
                                            .values(kp_data)
                                            .where(keypairs.c.user == o.uuid))
                        await graph_ctx.db_conn.execute(query)
                else:
                    # User becomes non-admin. Make the keypair non-admin as well.
                    # If there are multiple admin keypairs, inactivate them.
                    rows = result.fetchall()
                    cnt = 0
                    for row in rows:
                        kp_data = dict()
                        if cnt == 0:
                            kp_data['is_admin'] = False
                        elif row.is_admin and row.is_active:
                            kp_data['is_active'] = False
                        if len(kp_data.keys()) > 0:
                            query = (keypairs.update()
                                                .values(kp_data)
                                                .where(keypairs.c.user == row.user))
                            await graph_ctx.db_conn.execute(query)
                        cnt += 1

            # If domain is changed and no group is associated, clear previous domain's group.
            if prev_domain_name != o.domain_name and not props.group_ids:
                from .group import association_groups_users, groups
                query = (association_groups_users
                            .delete()
                            .where(association_groups_users.c.user_id == o.uuid))
                await graph_ctx.db_conn.execute(query)

            # Update user's group if group_ids parameter is provided.
            if props.group_ids and o is not None:
                from .group import association_groups_users, groups  # noqa
                # Clear previous groups associated with the user.
                query = (association_groups_users
                            .delete()
                            .where(association_groups_users.c.user_id == o.uuid))
                await graph_ctx.db_conn.execute(query)
                # Add user to new groups.
                query = (sa.select([groups.c.id])
                            .select_from(groups)
                            .where(groups.c.domain_name == o.domain_name)
                            .where(groups.c.id.in_(props.group_ids)))
                result = await graph_ctx.db_conn.execute(query)
                grps = result.fetchall()
                if grps:
                    values = [{'user_id': o.uuid, 'group_id': grp.id} for grp in grps]
                    query = association_groups_users.insert().values(values)
                    await graph_ctx.db_conn.execute(query)
            return cls(ok=True, msg='success', user=o)
        except sa.exc.IntegrityError as e:
            return cls(ok=False, msg=f'integrity error: {e}', user=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as e:
            return cls(ok=False, msg=f'unexpected error: {e}', user=None)


class DeleteUser(graphene.Mutation):
    """
    Instead of really deleting user, just mark the account as deleted status.

    All related keypairs will also be inactivated.
    """

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        email = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        email: str,
    ) -> DeleteUser:
        graph_ctx: GraphQueryContext = info.context
        try:
            # Make all user keypairs inactive.
            from ai.backend.manager.models import keypairs
            query = (
                keypairs.update()
                .values(is_active=False)
                .where(keypairs.c.user_id == email)
            )
            await graph_ctx.db_conn.execute(query)
            # Mark user as deleted.
            query = (
                users.update()
                .values(status=UserStatus.DELETED,
                        status_info='admin-requested')
                .where(users.c.email == email)
            )
            result = await graph_ctx.db_conn.execute(query)
            if result.rowcount > 0:
                return cls(ok=True, msg='success')
            else:
                return cls(ok=False, msg='no such user')
        except sa.exc.IntegrityError as e:
            return cls(ok=False, msg=f'integrity error: {e}')
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as e:
            return cls(ok=False, msg=f'unexpected error: {e}')


class PurgeUser(graphene.Mutation):
    """
    Delete user as well as all user-related DB informations such as keypairs, kernels, etc.

    If target user has virtual folders, they can be purged together or migrated to the superadmin.

    vFolder treatment policy:
      User-type:
      - vfolder is not shared: delete
      - vfolder is shared:
        + if purge_shared_vfolder is True: delete
        + else: change vfolder's owner to requested admin

    This action cannot be undone.
    """
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        email = graphene.String(required=True)
        props = PurgeUserInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        email: str,
        props: PurgeUserInput,
    ) -> PurgeUser:
        graph_ctx: GraphQueryContext = info.context
        try:
            query = (
                sa.select([users.c.uuid])
                .select_from(users)
                .where(users.c.email == email)
            )
            user_uuid = await graph_ctx.db_conn.scalar(query)
            log.info('completly deleting user {0}...', email)

            if await cls.user_vfolder_mounted_to_active_kernels(graph_ctx.db_conn, user_uuid):
                raise RuntimeError('Some of user\'s virtual folders are mounted to active kernels. '
                                    'Terminate those kernels first.')
            if await cls.user_has_active_kernels(graph_ctx.db_conn, user_uuid):
                raise RuntimeError('User has some active kernels. Terminate them first.')

            if not props.purge_shared_vfolders:
                await cls.migrate_shared_vfolders(
                    graph_ctx.db_conn,
                    deleted_user_uuid=user_uuid,
                    target_user_uuid=graph_ctx.user['uuid'],
                    target_user_email=graph_ctx.user['email'],
                )
            await cls.delete_vfolders(graph_ctx.db_conn, user_uuid, graph_ctx.storage_manager)
            await cls.delete_kernels(graph_ctx.db_conn, user_uuid)
            await cls.delete_keypairs(graph_ctx.db_conn, user_uuid)

            query = (
                users.delete()
                .where(users.c.email == email)
            )
            result = await graph_ctx.db_conn.execute(query)
            if result.rowcount > 0:
                log.info('user is deleted: {0}', email)
                return cls(ok=True, msg='success')
            else:
                return cls(ok=False, msg='no such user')
        except sa.exc.IntegrityError as e:
            return cls(ok=False, msg=f'integrity error: {e}')
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as e:
            return cls(ok=False, msg=f'unexpected error: {e}')

    @classmethod
    async def migrate_shared_vfolders(
        cls,
        conn: SAConnection,
        deleted_user_uuid: UUID,
        target_user_uuid: UUID,
        target_user_email: str,
    ) -> int:
        """
        Migrate shared virtual folders' ownership to a target user.

        If migrating virtual folder's name collides with target user's already
        existing folder, append random string to the migrating one.

        :param conn: DB connection
        :param deleted_user_uuid: user's UUID who will be deleted
        :param target_user_uuid: user's UUID who will get the ownership of virtual folders

        :return: number of deleted rows
        """
        from . import vfolders, vfolder_invitations, vfolder_permissions
        # Gather target user's virtual folders' names.
        query = (
            sa.select([vfolders.c.name])
            .select_from(vfolders)
            .where(vfolders.c.user == target_user_uuid)
        )
        existing_vfolder_names = [row.name async for row in (await conn.stream(query))]

        # Migrate shared virtual folders.
        # If virtual folder's name collides with target user's folder,
        # append random string to the name of the migrating folder.
        j = vfolder_permissions.join(
            vfolders,
            vfolder_permissions.c.vfolder == vfolders.c.id
        )
        query = (
            sa.select([vfolders.c.id, vfolders.c.name])
            .select_from(j)
            .where(vfolders.c.user == deleted_user_uuid)
        )
        migrate_updates = []
        async for row in (await conn.stream(query)):
            name = row.name
            if name in existing_vfolder_names:
                name += f'-{uuid4().hex[:10]}'
            migrate_updates.append({'vid': row.id, 'vname': name})
        if migrate_updates:
            # Remove invitations and vfolder_permissions from target user.
            # Target user will be the new owner, and it does not make sense to have
            # invitation and shared permission for its own folder.
            migrate_vfolder_ids = [item['vid'] for item in migrate_updates]
            query = (
                vfolder_invitations.delete()
                .where((vfolder_invitations.c.invitee == target_user_email) &
                       (vfolder_invitations.c.vfolder.in_(migrate_vfolder_ids)))
            )
            await conn.execute(query)
            query = (
                vfolder_permissions.delete()
                .where((vfolder_permissions.c.user == target_user_uuid) &
                       (vfolder_permissions.c.vfolder.in_(migrate_vfolder_ids)))
            )
            await conn.execute(query)

            rowcount = 0
            for item in migrate_updates:
                query = (
                    vfolders.update()
                    .values(
                        user=target_user_uuid,
                        name=item['vname'],
                    )
                    .where(vfolders.c.id == item['vid'])
                )
                result = await conn.execute(query)
                rowcount += result.rowcount
            if rowcount > 0:
                log.info('{0} shared folders detected. migrated to user {1}',
                         rowcount, target_user_uuid)
            return rowcount
        else:
            return 0

    @classmethod
    async def delete_vfolders(
        cls,
        conn: SAConnection,
        user_uuid: UUID,
        storage_manager: StorageSessionManager,
    ) -> int:
        """
        Delete user's all virtual folders as well as their physical data.

        :param conn: DB connection
        :param user_uuid: user's UUID to delete virtual folders

        :return: number of deleted rows
        """
        from . import vfolders, vfolder_permissions
        query = (
            vfolder_permissions.delete()
            .where(vfolder_permissions.c.user == user_uuid)
        )
        await conn.execute(query)
        query = (
            sa.select([vfolders.c.id, vfolders.c.host])
            .select_from(vfolders)
            .where(vfolders.c.user == user_uuid)
        )
        result = await conn.execute(query)
        target_vfs = result.fetchall()
        query = (vfolders.delete().where(vfolders.c.user == user_uuid))
        result = await conn.execute(query)
        for row in target_vfs:
            try:
                async with storage_manager.request(
                    row['host'], 'POST', 'folder/delete',
                    json={
                        'volume': storage_manager.split_host(row['host'])[1],
                        'vfid': str(row['id']),
                    },
                    raise_for_status=True,
                ):
                    pass
            except aiohttp.ClientResponseError:
                log.error('error on deleting vfolder filesystem directory: {0}', row['id'])
                raise VFolderOperationFailed
        if result.rowcount > 0:
            log.info('deleted {0} user\'s virtual folders ({1})', result.rowcount, user_uuid)
        return result.rowcount

    @classmethod
    async def user_vfolder_mounted_to_active_kernels(
        cls,
        conn: SAConnection,
        user_uuid: UUID,
    ) -> bool:
        """
        Check if no active kernel is using the user's virtual folders.

        :param conn: DB connection
        :param user_uuid: user's UUID

        :return: True if a virtual folder is mounted to active kernels.
        """
        from . import kernels, vfolders, AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES
        query = (
            sa.select([vfolders.c.id])
            .select_from(vfolders)
            .where(vfolders.c.user == user_uuid)
        )
        result = await conn.execute(query)
        rows = result.fetchall()
        user_vfolder_ids = [row.id for row in rows]
        query = (
            sa.select([kernels.c.mounts])
            .select_from(kernels)
            .where(kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
        )
        async for row in (await conn.stream(query)):
            for _mount in row['mounts']:
                try:
                    vfolder_id = UUID(_mount[2])
                    if vfolder_id in user_vfolder_ids:
                        return True
                except Exception:
                    pass
        return False

    @classmethod
    async def user_has_active_kernels(
        cls,
        conn: SAConnection,
        user_uuid: UUID,
    ) -> bool:
        """
        Check if the user does not have active kernels.

        :param conn: DB connection
        :param user_uuid: user's UUID

        :return: True if the user has some active kernels.
        """
        from . import kernels, AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES
        query = (
            sa.select([sa.func.count()])
            .select_from(kernels)
            .where((kernels.c.user_uuid == user_uuid) &
                   (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)))
        )
        active_kernel_count = await conn.scalar(query)
        return (active_kernel_count > 0)

    @classmethod
    async def delete_kernels(
        cls,
        conn: SAConnection,
        user_uuid: UUID,
    ) -> int:
        """
        Delete user's all kernels.

        :param conn: DB connection
        :param user_uuid: user's UUID to delete kernels
        :return: number of deleted rows
        """
        from . import kernels
        query = (
            kernels.delete()
            .where(kernels.c.user_uuid == user_uuid)
        )
        result = await conn.execute(query)
        if result.rowcount > 0:
            log.info('deleted {0} user\'s kernels ({1})', result.rowcount, user_uuid)
        return result.rowcount

    @classmethod
    async def delete_keypairs(
        cls,
        conn: SAConnection,
        user_uuid: UUID,
    ) -> int:
        """
        Delete user's all keypairs.

        :param conn: DB connection
        :param user_uuid: user's UUID to delete keypairs
        :return: number of deleted rows
        """
        from . import keypairs
        query = (
            keypairs.delete()
            .where(keypairs.c.user == user_uuid)
        )
        result = await conn.execute(query)
        if result.rowcount > 0:
            log.info('deleted {0} user\'s keypairs ({1})', result.rowcount, user_uuid)
        return result.rowcount


def _hash_password(password):
    return bcrypt.using(rounds=12).hash(password)


def _verify_password(guess, hashed):
    return bcrypt.verify(guess, hashed)


async def check_credential(
    db: SAEngine,
    domain: str,
    email: str,
    password: str,
) -> Any:
    async with db.begin() as conn:
        query = (
            sa.select([users])
            .select_from(users)
            .where(
                (users.c.email == email) &
                (users.c.domain_name == domain)
            )
        )
        result = await conn.execute(query)
        row = result.first()
        if row is None:
            return None
        if row['password'] is None:
            # user password is not set.
            return None
        try:
            if _verify_password(password, row['password']):
                return row
        except ValueError:
            return None
        return None



"""add vfolder_permissions table

Revision ID: 185852ff9872
Revises: 1fa6a31ea8e3
Create Date: 2018-07-05 16:02:05.225094

"""
from alembic import op
import sqlalchemy as sa
from ai.backend.manager.models.base import GUID


# revision identifiers, used by Alembic.
revision = '185852ff9872'
down_revision = '1fa6a31ea8e3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'vfolder_permissions',
        sa.Column('permission', sa.String(length=2), nullable=True),
        sa.Column('vfolder', GUID(), nullable=False),
        sa.Column('access_key', sa.String(length=20), nullable=False),
        sa.ForeignKeyConstraint(['access_key'], ['keypairs.access_key'],
                                name=op.f('fk_vfolder_permissions_access_key_keypairs')),
        sa.ForeignKeyConstraint(['vfolder'], ['vfolders.id'],
                                name=op.f('fk_vfolder_permissions_vfolder_vfolders'),
                                onupdate='CASCADE', ondelete='CASCADE')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('vfolder_permissions')
    # ### end Alembic commands ###



"""Enlarge kernels.lang column length

Revision ID: 10e39a34eed5
Revises: d582942886ad
Create Date: 2018-10-29 13:52:10.583443

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '10e39a34eed5'
down_revision = 'd582942886ad'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('kernels', 'lang',
                    existing_type=sa.String(length=64),
                    type_=sa.String(length=512))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('kernels', 'lang',
                    existing_type=sa.String(length=512),
                    type_=sa.String(length=64))
    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
import collections
import enum
import functools
import logging
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Sequence,
    TYPE_CHECKING,
    Type,
    TypeVar,
    Union,
)
import sys
import uuid

from aiodataloader import DataLoader
from aiotools import apartial
import graphene
from graphene.types import Scalar
from graphql.language import ast
from graphene.types.scalars import MIN_INT, MAX_INT
import sqlalchemy as sa
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import (
    AsyncConnection as SAConnection,
    AsyncEngine as SAEngine,
)
from sqlalchemy.types import (
    SchemaType,
    TypeDecorator,
    CHAR
)
from sqlalchemy.dialects.postgresql import UUID, ENUM, JSONB

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    BinarySize,
    KernelId,
    ResourceSlot,
    SessionId,
)

from .. import models
from ..api.exceptions import (
    GenericForbidden, InvalidAPIParameters,
)
if TYPE_CHECKING:
    from graphql.execution.executors.asyncio import AsyncioExecutor

    from .gql import GraphQueryContext
    from .user import UserRole

SAFE_MIN_INT = -9007199254740991
SAFE_MAX_INT = 9007199254740991

log = BraceStyleAdapter(logging.getLogger(__name__))

# The common shared metadata instance
convention = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = sa.MetaData(naming_convention=convention)

pgsql_connect_opts = {
    'server_settings': {'jit': 'off'}
}


# helper functions
def zero_if_none(val):
    return 0 if val is None else val


class EnumType(TypeDecorator, SchemaType):
    """
    A stripped-down version of Spoqa's sqlalchemy-enum34.
    It also handles postgres-specific enum type creation.

    The actual postgres enum choices are taken from the Python enum names.
    """

    impl = ENUM
    cache_ok = True

    def __init__(self, enum_cls, **opts):
        assert issubclass(enum_cls, enum.Enum)
        if 'name' not in opts:
            opts['name'] = enum_cls.__name__.lower()
        self._opts = opts
        enums = (m.name for m in enum_cls)
        super().__init__(*enums, **opts)
        self._enum_cls = enum_cls

    def process_bind_param(self, value, dialect):
        return value.name if value else None

    def process_result_value(self, value: str, dialect):
        return self._enum_cls[value] if value else None

    def copy(self):
        return EnumType(self._enum_cls, **self._opts)

    @property
    def python_type(self):
        return self._enum_class


class EnumValueType(TypeDecorator, SchemaType):
    """
    A stripped-down version of Spoqa's sqlalchemy-enum34.
    It also handles postgres-specific enum type creation.

    The actual postgres enum choices are taken from the Python enum values.
    """

    impl = ENUM
    cache_ok = True

    def __init__(self, enum_cls, **opts):
        assert issubclass(enum_cls, enum.Enum)
        if 'name' not in opts:
            opts['name'] = enum_cls.__name__.lower()
        self._opts = opts
        enums = (m.value for m in enum_cls)
        super().__init__(*enums, **opts)
        self._enum_cls = enum_cls

    def process_bind_param(self, value, dialect):
        return value.value if value else None

    def process_result_value(self, value: str, dialect):
        return self._enum_cls(value) if value else None

    def copy(self):
        return EnumValueType(self._enum_cls, **self._opts)

    @property
    def python_type(self):
        return self._enum_class


class ResourceSlotColumn(TypeDecorator):
    """
    A column type wrapper for ResourceSlot from JSONB.
    """

    impl = JSONB
    cache_ok = True

    def process_bind_param(self, value: Union[Mapping, ResourceSlot], dialect):
        if isinstance(value, Mapping) and not isinstance(value, ResourceSlot):
            return value
        return value.to_json() if value is not None else None

    def process_result_value(self, raw_value: Dict[str, str], dialect):
        # legacy handling
        interim_value: Dict[str, Any] = raw_value
        mem = raw_value.get('mem')
        if isinstance(mem, str) and not mem.isdigit():
            interim_value['mem'] = BinarySize.from_str(mem)
        return ResourceSlot.from_json(interim_value) if raw_value is not None else None

    def copy(self):
        return ResourceSlotColumn()


class CurrencyTypes(enum.Enum):
    KRW = 'KRW'
    USD = 'USD'


UUID_SubType = TypeVar('UUID_SubType', bound=uuid.UUID)


class GUID(TypeDecorator, Generic[UUID_SubType]):
    """
    Platform-independent GUID type.
    Uses PostgreSQL's UUID type, otherwise uses CHAR(16) storing as raw bytes.
    """
    impl = CHAR
    uuid_subtype_func: ClassVar[Callable[[Any], UUID_SubType]] = lambda v: v
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(16))

    def process_bind_param(self, value: Union[UUID_SubType, uuid.UUID], dialect):
        # NOTE: SessionId, KernelId are *not* actual types defined as classes,
        #       but a "virtual" type that is an identity function at runtime.
        #       The type checker treats them as distinct derivatives of uuid.UUID.
        #       Therefore, we just do isinstance on uuid.UUID only below.
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            if isinstance(value, uuid.UUID):
                return str(value)
            else:
                return str(uuid.UUID(value))
        else:
            if isinstance(value, uuid.UUID):
                return value.bytes
            else:
                return uuid.UUID(value).bytes

    def process_result_value(self, value: Any, dialect) -> Optional[UUID_SubType]:
        if value is None:
            return value
        else:
            cls = type(self)
            return cls.uuid_subtype_func(uuid.UUID(value))


class SessionIDColumnType(GUID):
    uuid_subtype_func = SessionId


class KernelIDColumnType(GUID):
    uuid_subtype_func = KernelId


def IDColumn(name='id'):
    return sa.Column(name, GUID, primary_key=True,
                     server_default=sa.text("uuid_generate_v4()"))


def SessionIDColumn(name='id'):
    return sa.Column(name, SessionIDColumnType, primary_key=True,
                     server_default=sa.text("uuid_generate_v4()"))


def KernelIDColumn(name='id'):
    return sa.Column(name, KernelIDColumnType, primary_key=True,
                     server_default=sa.text("uuid_generate_v4()"))


def ForeignKeyIDColumn(name, fk_field, nullable=True):
    return sa.Column(name, GUID, sa.ForeignKey(fk_field), nullable=nullable)


class DataLoaderManager:
    """
    For every different combination of filtering conditions, we need to make a
    new DataLoader instance because it "batches" the database queries.
    This manager get-or-creates dataloaders with fixed conditions (represetned
    as arguments) like a cache.

    NOTE: Just like DataLoaders, it is recommended to instantiate this manager
    for every incoming API request.
    """

    cache: Dict[int, DataLoader]

    def __init__(self) -> None:
        self.cache = {}
        self.mod = sys.modules['ai.backend.manager.models']

    @staticmethod
    def _get_key(otname: str, args, kwargs) -> int:
        """
        Calculate the hash of the all arguments and keyword arguments.
        """
        key = (otname, ) + args
        for item in kwargs.items():
            key += item
        return hash(key)

    def get_loader(self, context: GraphQueryContext, objtype_name: str, *args, **kwargs) -> DataLoader:
        k = self._get_key(objtype_name, args, kwargs)
        loader = self.cache.get(k)
        if loader is None:
            objtype_name, has_variant, variant_name = objtype_name.partition('.')
            objtype = getattr(self.mod, objtype_name)
            if has_variant:
                batch_load_fn = getattr(objtype, 'batch_load_' + variant_name)
            else:
                batch_load_fn = objtype.batch_load
            loader = DataLoader(
                apartial(batch_load_fn, context, *args, **kwargs),
                max_batch_size=128,
            )
            self.cache[k] = loader
        return loader


class ResourceLimit(graphene.ObjectType):
    key = graphene.String()
    min = graphene.String()
    max = graphene.String()


class KVPair(graphene.ObjectType):
    key = graphene.String()
    value = graphene.String()


class BigInt(Scalar):
    """
    BigInt is an extension of the regular graphene.Int scalar type
    to support integers outside the range of a signed 32-bit integer.
    """

    @staticmethod
    def coerce_bigint(value):
        num = int(value)
        if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
            raise ValueError(
                'Cannot serialize integer out of the safe range.')
        if not (MIN_INT <= num <= MAX_INT):
            # treat as float
            return float(int(num))
        return num

    serialize = coerce_bigint
    parse_value = coerce_bigint

    @staticmethod
    def parse_literal(node):
        if isinstance(node, ast.IntValue):
            num = int(node.value)
            if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
                raise ValueError(
                    'Cannot parse integer out of the safe range.')
            if not (MIN_INT <= num <= MAX_INT):
                # treat as float
                return float(int(num))
            return num


class Item(graphene.Interface):
    id = graphene.ID()


class PaginatedList(graphene.Interface):
    items = graphene.List(Item, required=True)
    total_count = graphene.Int(required=True)


# ref: https://github.com/python/mypy/issues/1212
_GenericSQLBasedGQLObject = TypeVar('_GenericSQLBasedGQLObject', bound='_SQLBasedGQLObject')
_Key = TypeVar('_Key')


class _SQLBasedGQLObject(Protocol):
    @classmethod
    def from_row(
        cls: Type[_GenericSQLBasedGQLObject],
        ctx: GraphQueryContext,
        row: Row,
    ) -> _GenericSQLBasedGQLObject:
        ...


async def batch_result(
    graph_ctx: GraphQueryContext,
    db_conn: SAConnection,
    query: sa.sql.Select,
    obj_type: Type[_GenericSQLBasedGQLObject],
    key_list: Iterable[_Key],
    key_getter: Callable[[Row], _Key],
) -> Sequence[Optional[_GenericSQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    """
    objs_per_key: Dict[_Key, Optional[_GenericSQLBasedGQLObject]]
    objs_per_key = collections.OrderedDict()
    for key in key_list:
        objs_per_key[key] = None
    async for row in (await db_conn.stream(query)):
        objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_multiresult(
    graph_ctx: GraphQueryContext,
    db_conn: SAConnection,
    query: sa.sql.Select,
    obj_type: Type[_GenericSQLBasedGQLObject],
    key_list: Iterable[_Key],
    key_getter: Callable[[Row], _Key],
) -> Sequence[Sequence[_GenericSQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    """
    objs_per_key: Dict[_Key, List[_GenericSQLBasedGQLObject]]
    objs_per_key = collections.OrderedDict()
    for key in key_list:
        objs_per_key[key] = list()
    async for row in (await db_conn.stream(query)):
        objs_per_key[key_getter(row)].append(
            obj_type.from_row(graph_ctx, row)
        )
    return [*objs_per_key.values()]


def privileged_query(required_role: UserRole):

    def wrap(func):

        @functools.wraps(func)
        async def wrapped(executor: AsyncioExecutor, info: graphene.ResolveInfo, *args, **kwargs) -> Any:
            from .user import UserRole
            ctx: GraphQueryContext = info.context
            if ctx.user['role'] != UserRole.SUPERADMIN:
                raise GenericForbidden('superadmin privilege required')
            return await func(executor, info, *args, **kwargs)

        return wrapped

    return wrap


def scoped_query(
    *,
    autofill_user: bool = False,
    user_key: str = 'access_key',
):
    """
    Prepends checks for domain/group/user access rights depending
    on the client's user and keypair information.

    :param autofill_user: When the *user_key* is not specified,
        automatically fills out the user data with the current
        user who is makeing the API request.
    :param user_key: The key used for storing user identification value
        in the keyword arguments.
    """

    def wrap(resolve_func):

        @functools.wraps(resolve_func)
        async def wrapped(executor: AsyncioExecutor, info: graphene.ResolveInfo, *args, **kwargs) -> Any:
            from .user import UserRole
            ctx: GraphQueryContext = info.context
            client_role = ctx.user['role']
            if user_key == 'access_key':
                client_user_id = ctx.access_key
            elif user_key == 'email':
                client_user_id = ctx.user['email']
            else:
                client_user_id = ctx.user['uuid']
            client_domain = ctx.user['domain_name']
            domain_name = kwargs.get('domain_name', None)
            group_id = kwargs.get('group_id', None)
            user_id = kwargs.get(user_key, None)
            if client_role == UserRole.SUPERADMIN:
                if autofill_user:
                    if user_id is None:
                        user_id = client_user_id
            elif client_role == UserRole.ADMIN:
                if domain_name is not None and domain_name != client_domain:
                    raise GenericForbidden
                domain_name = client_domain
                if group_id is not None:
                    # TODO: check if the group is a member of the domain
                    pass
                if autofill_user:
                    if user_id is None:
                        user_id = client_user_id
            elif client_role == UserRole.USER:
                if domain_name is not None and domain_name != client_domain:
                    raise GenericForbidden
                domain_name = client_domain
                if group_id is not None:
                    # TODO: check if the group is a member of the domain
                    # TODO: check if the client is a member of the group
                    pass
                if user_id is not None and user_id != client_user_id:
                    raise GenericForbidden
                user_id = client_user_id
            else:
                raise InvalidAPIParameters('Unknown client role')
            kwargs['domain_name'] = domain_name
            if group_id is not None:
                kwargs['group_id'] = group_id
            kwargs[user_key] = user_id
            return await resolve_func(executor, info, *args, **kwargs)

        return wrapped

    return wrap


def privileged_mutation(required_role, target_func=None):

    def wrap(func):

        @functools.wraps(func)
        async def wrapped(cls, root, info: graphene.ResolveInfo, *args, **kwargs) -> Any:
            from .user import UserRole
            from .group import groups  # , association_groups_users
            ctx: GraphQueryContext = info.context
            permitted = False
            if required_role == UserRole.SUPERADMIN:
                if ctx.user['role'] == required_role:
                    permitted = True
            elif required_role == UserRole.ADMIN:
                if ctx.user['role'] == UserRole.SUPERADMIN:
                    permitted = True
                elif ctx.user['role'] == UserRole.USER:
                    permitted = False
                else:
                    if target_func is None:
                        return cls(False, 'misconfigured privileged mutation: no target_func', None)
                    target_domain, target_group = target_func(*args, **kwargs)
                    if target_domain is None and target_group is None:
                        return cls(False, 'misconfigured privileged mutation: '
                                          'both target_domain and target_group missing', None)
                    permit_chains = []
                    if target_domain is not None:
                        if ctx.user['domain_name'] == target_domain:
                            permit_chains.append(True)
                    if target_group is not None:
                        async with ctx.db.begin() as conn:
                            # check if the group is part of the requester's domain.
                            query = (
                                groups.select()
                                .where(
                                    (groups.c.id == target_group) &
                                    (groups.c.domain_name == ctx.user['domain_name'])
                                )
                            )
                            result = await conn.execute(query)
                            if result.rowcount > 0:
                                permit_chains.append(True)
                            # TODO: check the group permission if implemented
                            # query = (
                            #     association_groups_users.select()
                            #     .where(association_groups_users.c.group_id == target_group)
                            # )
                            # result = await conn.execute(query)
                            # if result.rowcount > 0:
                            #     permit_chains.append(True)
                    permitted = all(permit_chains) if permit_chains else False
            elif required_role == UserRole.USER:
                permitted = True
            # assuming that mutation result objects has 2 or 3 fields:
            # success(bool), message(str) - usually for delete mutations
            # success(bool), message(str), item(object)
            if permitted:
                return await func(cls, root, info, *args, **kwargs)
            return cls(False, f"no permission to execute {info.path[0]}")

        return wrapped

    return wrap


ResultType = TypeVar('ResultType', bound=graphene.ObjectType)
ItemType = TypeVar('ItemType', bound=graphene.ObjectType)


async def simple_db_mutate(
    result_cls: Type[ResultType],
    graph_ctx: GraphQueryContext,
    mutation_query: sa.sql.Update | sa.sql.Insert,
) -> ResultType:
    try:
        result = await graph_ctx.db_conn.execute(mutation_query)
        if result.rowcount > 0:
            return result_cls(True, 'success')
        else:
            return result_cls(False, 'no matching record')
    except sa.exc.IntegrityError as e:
        return result_cls(False, f'integrity error: {e}')
    except (asyncio.CancelledError, asyncio.TimeoutError):
        raise
    except Exception as e:
        return result_cls(False, f'unexpected error: {e}')


async def simple_db_mutate_returning_item(
    result_cls: Type[ResultType],
    graph_ctx: GraphQueryContext,
    mutation_query: sa.sql.Update | sa.sql.Insert,
    *,
    item_query: sa.sql.Select,
    item_cls: Type[ItemType],
) -> ResultType:
    try:
        result = await graph_ctx.db_conn.execute(mutation_query)
        if result.rowcount > 0:
            result = await graph_ctx.db_conn.execute(item_query)
            item = result.first()
            return result_cls(True, 'success', item_cls.from_row(graph_ctx, item))
        else:
            return result_cls(False, 'no matching record', None)
    except sa.exc.IntegrityError as e:
        return result_cls(False, f'integrity error: {e}', None)
    except (asyncio.CancelledError, asyncio.TimeoutError):
        raise
    except Exception as e:
        return result_cls(False, f'unexpected error: {e}', None)


def set_if_set(src: object, target: MutableMapping[str, Any], name: str, *, clean_func=None) -> None:
    v = getattr(src, name)
    # NOTE: unset optional fields are passed as null.
    if v is not None:
        if callable(clean_func):
            target[name] = clean_func(v)
        else:
            target[name] = v


async def populate_fixture(
    engine: SAEngine,
    fixture_data: Mapping[str, Sequence[Dict[str, Any]]],
    *,
    ignore_unique_violation: bool = False,
) -> None:
    for table_name, rows in fixture_data.items():
        table: sa.Table = getattr(models, table_name)
        assert isinstance(table, sa.Table)
        async with engine.begin() as conn:
            for col in table.columns:
                if isinstance(col.type, EnumType):
                    for row in rows:
                        row[col.name] = col.type._enum_cls[row[col.name]]
                elif isinstance(col.type, EnumValueType):
                    for row in rows:
                        row[col.name] = col.type._enum_cls(row[col.name])
            await conn.execute(sa.dialects.postgresql.insert(table, rows).on_conflict_do_nothing())



from __future__ import annotations

from contextvars import ContextVar
from datetime import datetime
import logging
import pkg_resources
from typing import (
    Any,
    Awaitable,
    Final,
    List,
    Mapping,
    Sequence,
    Tuple,
    Union,
    TYPE_CHECKING,
)

import aioredis
import aiotools
from dateutil.tz import tzutc
import sqlalchemy as sa
from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import (
    AsyncConnection as SAConnection,
    AsyncEngine as SAEngine,
)
from sqlalchemy.sql.expression import true

from ai.backend.common.events import (
    AgentStartedEvent,
    CoalescingOptions,
    DoScheduleEvent,
    DoPrepareEvent,
    SessionCancelledEvent,
    SessionEnqueuedEvent,
    SessionPreparingEvent,
    SessionScheduledEvent,
    SessionTerminatedEvent,
    EventDispatcher,
    EventProducer,
)
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    aobject,
    AgentId,
    ClusterMode,
    ResourceSlot,
)

from ..api.exceptions import InstanceNotAvailable
from ..distributed import GlobalTimer
from ..defs import (
    AdvisoryLock,
    REDIS_STREAM_DB,
)
from ..exceptions import convert_to_status_data
from ..models import (
    agents, kernels, scaling_groups,
    recalc_agent_resource_occupancy,
    recalc_concurrency_used,
    AgentStatus, KernelStatus,
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
)
from ..models.utils import (
    execute_with_retry,
    sql_json_increment,
    sql_json_merge,
)
from .types import (
    PredicateResult,
    PendingSession,
    ExistingSession,
    SchedulingContext,
    AgentContext,
    AgentAllocationContext,
    AbstractScheduler,
    KernelAgentBinding,
)
from .predicates import (
    check_reserved_batch_session,
    check_concurrency,
    check_dependencies,
    check_keypair_resource_limit,
    check_group_resource_limit,
    check_domain_resource_limit,
    check_scaling_group,
)

if TYPE_CHECKING:
    from ..config import LocalConfig, SharedConfig
    from ..registry import AgentRegistry

__all__ = (
    'load_scheduler',
    'SchedulerDispatcher',
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.manager.scheduler'))

_log_fmt: ContextVar[str] = ContextVar('_log_fmt')
_log_args: ContextVar[Tuple[Any, ...]] = ContextVar('_log_args')

_key_schedule_prep_tasks: Final = "scheduler.preptasks"


def load_scheduler(name: str, scheduler_configs: Mapping[str, Any]) -> AbstractScheduler:
    entry_prefix = 'backendai_scheduler_v10'
    for entrypoint in pkg_resources.iter_entry_points(entry_prefix):
        if entrypoint.name == name:
            log.debug('loading scheduler plugin "{}" from {}', name, entrypoint.module_name)
            scheduler_cls = entrypoint.load()
            scheduler_config = scheduler_configs.get(name, {})
            return scheduler_cls(scheduler_config)
    raise ImportError('Cannot load the scheduler plugin', name)


StartTaskArgs = Tuple[
    Tuple[Any, ...],
    SchedulingContext,
    Tuple[PendingSession, List[KernelAgentBinding]],
    List[Tuple[str, Union[Exception, PredicateResult]]],
]


class SchedulerDispatcher(aobject):

    config: LocalConfig
    shared_config: SharedConfig
    registry: AgentRegistry
    db: SAEngine

    schedule_timer_redis: aioredis.Redis
    prepare_timer_redis: aioredis.Redis
    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    schedule_timer: GlobalTimer
    prepare_timer: GlobalTimer

    def __init__(
        self,
        local_config: LocalConfig,
        shared_config: SharedConfig,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        registry: AgentRegistry,
    ) -> None:
        self.local_config = local_config
        self.shared_config = shared_config
        self.event_dispatcher = event_dispatcher
        self.event_producer = event_producer
        self.registry = registry
        self.db = registry.db

    async def __ainit__(self) -> None:
        coalescing_opts: CoalescingOptions = {
            'max_wait': 0.5,
            'max_batch_size': 32,
        }
        # coalescing_opts = None
        evd = self.registry.event_dispatcher
        evd.consume(SessionEnqueuedEvent, None, self.schedule, coalescing_opts, name="dispatcher.enq")
        evd.consume(SessionTerminatedEvent, None, self.schedule, coalescing_opts, name="dispatcher.term")
        evd.consume(AgentStartedEvent, None, self.schedule)
        evd.consume(DoScheduleEvent, None, self.schedule, coalescing_opts)
        evd.consume(DoPrepareEvent, None, self.prepare)
        redis_url = self.shared_config.get_redis_url(db=REDIS_STREAM_DB)
        self.schedule_timer_redis = await aioredis.create_redis(str(redis_url))
        self.prepare_timer_redis = await aioredis.create_redis(str(redis_url))
        self.schedule_timer = GlobalTimer(
            self.schedule_timer_redis,
            "scheduler_tick",
            self.event_producer,
            lambda: DoScheduleEvent(),
            interval=10.0,
        )
        self.prepare_timer = GlobalTimer(
            self.prepare_timer_redis,
            "prepare_tick",
            self.event_producer,
            lambda: DoPrepareEvent(),
            interval=10.0,
            initial_delay=5.0,
        )
        await self.schedule_timer.join()
        await self.prepare_timer.join()
        log.info('Session scheduler started')

    async def close(self) -> None:
        await self.prepare_timer.leave()
        await self.schedule_timer.leave()
        log.info('Session scheduler stopped')
        self.prepare_timer_redis.close()
        self.schedule_timer_redis.close()
        await self.prepare_timer_redis.wait_closed()
        await self.schedule_timer_redis.wait_closed()

    async def schedule(
        self,
        context: None,
        source: AgentId,
        event: SessionEnqueuedEvent | SessionTerminatedEvent | AgentStartedEvent | DoScheduleEvent,
    ) -> None:
        """
        Trigger the scheduler to scan pending sessions and mark them scheduled if they fulfill
        the scheduling requirements.

        HoL blocking issue due to indefinitely preparing sessions will be mitigated because
        they will be treated as already "scheduled" sessions and the scheduler will continue to
        work on other pending sessions.

        Session status transition: PENDING -> SCHEDULED
        """
        log.debug('schedule(): triggered')
        known_slot_types = await self.shared_config.get_resource_slots()
        sched_ctx = SchedulingContext(
            registry=self.registry,
            known_slot_types=known_slot_types,
        )

        # We use short transaction blocks to prevent deadlock timeouts under heavy loads
        # because this scheduling handler will be executed by only one process.
        # It is executed under a globally exclusive context using aioredlock.
        async with self.db.advisory_lock(AdvisoryLock.LOCKID_SCHEDULE):
            async with self.db.begin_readonly() as conn:
                query = (
                    sa.select([agents.c.scaling_group])
                    .select_from(agents)
                    .where(agents.c.status == AgentStatus.ALIVE)
                    .group_by(agents.c.scaling_group)
                )
                result = await conn.execute(query)
                schedulable_scaling_groups = [
                    row.scaling_group for row in result.fetchall()
                ]
            for sgroup_name in schedulable_scaling_groups:
                try:
                    await self._schedule_in_sgroup(
                        sched_ctx, sgroup_name,
                    )
                except InstanceNotAvailable:
                    # Proceed to the next scaling group and come back later.
                    log.debug('schedule({}): instance not available', sgroup_name)
                except Exception as e:
                    log.exception('schedule({}): scheduling error!\n{}', sgroup_name, repr(e))

    async def _load_scheduler(
        self,
        db_conn: SAConnection,
        sgroup_name: str,
    ) -> AbstractScheduler:
        query = (
            sa.select([scaling_groups.c.scheduler])
            .select_from(scaling_groups)
            .where(scaling_groups.c.name == sgroup_name)
        )
        result = await db_conn.execute(query)
        scheduler_name = result.scalar()
        return load_scheduler(scheduler_name, self.shared_config['plugins']['scheduler'])

    async def _schedule_in_sgroup(
        self,
        sched_ctx: SchedulingContext,
        sgroup_name: str,
    ) -> None:
        async with self.db.begin_readonly() as kernel_db_conn:
            scheduler = await self._load_scheduler(kernel_db_conn, sgroup_name)
            pending_sessions = await _list_pending_sessions(kernel_db_conn, sgroup_name)
            existing_sessions = await _list_existing_sessions(kernel_db_conn, sgroup_name)
        log.debug('running scheduler (sgroup:{}, pending:{}, existing:{})',
                  sgroup_name, len(pending_sessions), len(existing_sessions))
        zero = ResourceSlot()
        num_scheduled = 0
        while len(pending_sessions) > 0:

            async with self.db.begin() as conn:
                candidate_agents = await _list_agents_by_sgroup(conn, sgroup_name)
                total_capacity = sum((ag.available_slots for ag in candidate_agents), zero)

            picked_session_id = scheduler.pick_session(
                total_capacity,
                pending_sessions,
                existing_sessions,
            )
            if picked_session_id is None:
                # no session is picked.
                # continue to next sgroup.
                return
            for picked_idx, sess_ctx in enumerate(pending_sessions):
                if sess_ctx.session_id == picked_session_id:
                    break
            else:
                # no matching entry for picked session?
                raise RuntimeError('should not reach here')
            sess_ctx = pending_sessions.pop(picked_idx)

            log_fmt = 'schedule(s:{}, type:{}, name:{}, ak:{}, cluster_mode:{}): '
            log_args = (
                sess_ctx.session_id,
                sess_ctx.session_type,
                sess_ctx.session_name,
                sess_ctx.access_key,
                sess_ctx.cluster_mode,
            )
            _log_fmt.set(log_fmt)
            _log_args.set(log_args)
            log.debug(log_fmt + 'try-scheduling', *log_args)

            async def _check_predicates() -> List[Tuple[str, Union[Exception, PredicateResult]]]:
                check_results: List[Tuple[str, Union[Exception, PredicateResult]]] = []
                async with self.db.begin() as kernel_db_conn:
                    predicates: Sequence[Tuple[str, Awaitable[PredicateResult]]] = [
                        (
                            'reserved_time',
                            check_reserved_batch_session(kernel_db_conn, sched_ctx, sess_ctx),
                        ),
                        ('concurrency', check_concurrency(kernel_db_conn, sched_ctx, sess_ctx)),
                        ('dependencies', check_dependencies(kernel_db_conn, sched_ctx, sess_ctx)),
                        (
                            'keypair_resource_limit',
                            check_keypair_resource_limit(kernel_db_conn, sched_ctx, sess_ctx),
                        ),
                        (
                            'user_group_resource_limit',
                            check_group_resource_limit(kernel_db_conn, sched_ctx, sess_ctx),
                        ),
                        (
                            'domain_resource_limit',
                            check_domain_resource_limit(kernel_db_conn, sched_ctx, sess_ctx),
                        ),
                        (
                            'scaling_group_resource_limit',
                            check_scaling_group(kernel_db_conn, sched_ctx, sess_ctx),
                        ),
                    ]
                    for predicate_name, check_coro in predicates:
                        try:
                            check_results.append((predicate_name, await check_coro))
                        except DBAPIError:
                            raise
                        except Exception as e:
                            log.exception(log_fmt + 'predicate-error', *log_args)
                            check_results.append((predicate_name, e))
                return check_results

            check_results = await execute_with_retry(_check_predicates)
            has_failure = False
            has_permanent_failure = False
            failed_predicates = []
            passed_predicates = []
            for predicate_name, result in check_results:
                if isinstance(result, Exception):
                    has_failure = True
                    failed_predicates.append({
                        'name': predicate_name,
                        'msg': repr(result),
                    })
                    continue
                if result.passed:
                    passed_predicates.append({
                        'name': predicate_name,
                    })
                else:
                    failed_predicates.append({
                        'name': predicate_name,
                        'msg': result.message or "",
                    })
                    has_failure = True
                    if result.permanent:
                        has_permanent_failure = True  # noqa
            if has_failure:
                log.debug(log_fmt + 'predicate-checks-failed (temporary)', *log_args)
                # TODO: handle has_permanent_failure as cancellation
                #  - An early implementation of it has caused DB query blocking due to
                #    the inclusion of the kernels.status field. :(
                #    Let's fix it.

                async def _update() -> None:
                    async with self.db.begin() as conn:
                        await _rollback_predicate_mutations(
                            conn, sched_ctx, sess_ctx,
                        )
                        query = kernels.update().values({
                            'status_info': "predicate-checks-failed",
                            'status_data': sql_json_increment(
                                kernels.c.status_data,
                                ('scheduler', 'retries'),
                                parent_updates={
                                    'last_try': datetime.now(tzutc()).isoformat(),
                                    'failed_predicates': failed_predicates,
                                    'passed_predicates': passed_predicates,
                                }
                            ),
                        }).where(kernels.c.id == sess_ctx.session_id)
                        await conn.execute(query)

                await execute_with_retry(_update)
                # Predicate failures are *NOT* permanent errors.
                # We need to retry the scheduling afterwards.
                continue
            else:
                async def _update() -> None:
                    async with self.db.begin() as conn:
                        query = kernels.update().values({
                            'status_data': sql_json_merge(
                                kernels.c.status_data,
                                ('scheduler',),
                                {
                                    'last_try': datetime.now(tzutc()).isoformat(),
                                    'failed_predicates': failed_predicates,
                                    'passed_predicates': passed_predicates,
                                }
                            ),
                        }).where(kernels.c.id == sess_ctx.session_id)
                        await conn.execute(query)

                await execute_with_retry(_update)

            if sess_ctx.cluster_mode == ClusterMode.SINGLE_NODE:
                await self._schedule_single_node_session(
                    sched_ctx,
                    scheduler,
                    sgroup_name,
                    candidate_agents,
                    sess_ctx,
                    check_results,
                )
            elif sess_ctx.cluster_mode == ClusterMode.MULTI_NODE:
                await self._schedule_multi_node_session(
                    sched_ctx,
                    scheduler,
                    sgroup_name,
                    candidate_agents,
                    sess_ctx,
                    check_results,
                )
            else:
                raise RuntimeError(
                    f"should not reach here; unknown cluster_mode: {sess_ctx.cluster_mode}"
                )
            num_scheduled += 1
        if num_scheduled > 0:
            await self.event_producer.produce_event(DoPrepareEvent())

    async def _schedule_single_node_session(
        self,
        sched_ctx: SchedulingContext,
        scheduler: AbstractScheduler,
        sgroup_name: str,
        candidate_agents: Sequence[AgentContext],
        sess_ctx: PendingSession,
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        # Assign agent resource per session.
        log_fmt = _log_fmt.get()
        log_args = _log_args.get()
        try:
            agent_id = scheduler.assign_agent_for_session(candidate_agents, sess_ctx)
            if agent_id is None:
                raise InstanceNotAvailable
            async with self.db.begin() as agent_db_conn:
                agent_alloc_ctx = await _reserve_agent(
                    sched_ctx, agent_db_conn, sgroup_name, agent_id, sess_ctx.requested_slots,
                )
        except InstanceNotAvailable:
            log.debug(log_fmt + 'no-available-instances', *log_args)

            async def _update() -> None:
                async with self.db.begin() as kernel_db_conn:
                    await _rollback_predicate_mutations(
                        kernel_db_conn, sched_ctx, sess_ctx,
                    )
                    query = kernels.update().values({
                        'status_info': "no-available-instances",
                        'status_data': sql_json_increment(
                            kernels.c.status_data,
                            ('scheduler', 'retries'),
                            parent_updates={
                                'last_try': datetime.now(tzutc()).isoformat(),
                            }
                        ),
                    }).where(kernels.c.id == sess_ctx.session_id)
                    await kernel_db_conn.execute(query)

            await execute_with_retry(_update)
            raise
        except Exception as e:
            log.exception(
                log_fmt + 'unexpected-error, during agent allocation',
                *log_args,
            )
            exc_data = convert_to_status_data(e)

            async def _update() -> None:
                async with self.db.begin() as kernel_db_conn:
                    await _rollback_predicate_mutations(
                        kernel_db_conn, sched_ctx, sess_ctx,
                    )
                    query = kernels.update().values({
                        'status_info': "scheduler-error",
                        'status_data': exc_data,
                    }).where(kernels.c.id == sess_ctx.session_id)
                    await kernel_db_conn.execute(query)

            await execute_with_retry(_update)
            raise

        async def _finalize_scheduled() -> None:
            async with self.db.begin() as kernel_db_conn:
                query = kernels.update().values({
                    'agent': agent_alloc_ctx.agent_id,
                    'agent_addr': agent_alloc_ctx.agent_addr,
                    'scaling_group': sgroup_name,
                    'status': KernelStatus.SCHEDULED,
                    'status_info': 'scheduled',
                    'status_data': {},
                    'status_changed': datetime.now(tzutc()),
                }).where(kernels.c.session_id == sess_ctx.session_id)
                await kernel_db_conn.execute(query)

        await execute_with_retry(_finalize_scheduled)
        await self.registry.event_producer.produce_event(
            SessionScheduledEvent(sess_ctx.session_id, sess_ctx.session_creation_id)
        )

    async def _schedule_multi_node_session(
        self,
        sched_ctx: SchedulingContext,
        scheduler: AbstractScheduler,
        sgroup_name: str,
        candidate_agents: Sequence[AgentContext],
        sess_ctx: PendingSession,
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        # Assign agent resource per kernel in the session.
        log_fmt = _log_fmt.get()
        log_args = _log_args.get()
        agent_query_extra_conds = None
        kernel_agent_bindings: List[KernelAgentBinding] = []
        async with self.db.begin() as agent_db_conn:
            # This outer transaction is rolled back when any exception occurs inside,
            # including scheduling failures of a kernel.
            # It ensures that occupied_slots are recovered when there are partial
            # scheduling failures.
            for kernel in sess_ctx.kernels:
                try:
                    agent_alloc_ctx: AgentAllocationContext
                    agent_id = scheduler.assign_agent_for_kernel(candidate_agents, kernel)
                    if agent_id is None:
                        raise InstanceNotAvailable

                    async def _reserve() -> None:
                        nonlocal agent_alloc_ctx, candidate_agents
                        assert agent_id is not None
                        async with agent_db_conn.begin_nested():
                            agent_alloc_ctx = await _reserve_agent(
                                sched_ctx, agent_db_conn,
                                sgroup_name, agent_id, kernel.requested_slots,
                                extra_conds=agent_query_extra_conds,
                            )
                            candidate_agents = await _list_agents_by_sgroup(agent_db_conn, sgroup_name)

                    await execute_with_retry(_reserve)
                except InstanceNotAvailable:
                    log.debug(log_fmt + 'no-available-instances', *log_args)

                    async def _update() -> None:
                        async with self.db.begin() as kernel_db_conn:
                            await _rollback_predicate_mutations(
                                kernel_db_conn, sched_ctx, sess_ctx,
                            )
                            query = kernels.update().values({
                                'status_info': "no-available-instances",
                                'status_data': sql_json_increment(
                                    kernels.c.status_data,
                                    ('scheduler', 'retries'),
                                    parent_updates={
                                        'last_try': datetime.now(tzutc()).isoformat(),
                                    }
                                ),
                            }).where(kernels.c.id == kernel.kernel_id)
                            await kernel_db_conn.execute(query)

                    await execute_with_retry(_update)
                    raise
                except Exception as e:
                    log.exception(
                        log_fmt + 'unexpected-error, during agent allocation',
                        *log_args,
                    )
                    exc_data = convert_to_status_data(e)

                    async def _update() -> None:
                        async with self.db.begin() as kernel_db_conn:
                            await _rollback_predicate_mutations(
                                kernel_db_conn, sched_ctx, sess_ctx,
                            )
                            query = kernels.update().values({
                                'status_info': "scheduler-error",
                                'status_data': exc_data,
                            }).where(kernels.c.id == kernel.kernel_id)
                            await kernel_db_conn.execute(query)

                    await execute_with_retry(_update)
                    raise
                else:
                    kernel_agent_bindings.append(KernelAgentBinding(kernel, agent_alloc_ctx))

        assert len(kernel_agent_bindings) == len(sess_ctx.kernels)
        # Proceed to PREPARING only when all kernels are successfully scheduled.

        async def _finalize_scheduled() -> None:
            async with self.db.begin() as kernel_db_conn:
                for binding in kernel_agent_bindings:
                    query = kernels.update().values({
                        'agent': binding.agent_alloc_ctx.agent_id,
                        'agent_addr': binding.agent_alloc_ctx.agent_addr,
                        'scaling_group': sgroup_name,
                        'status': KernelStatus.SCHEDULED,
                        'status_info': 'scheduled',
                        'status_data': {},
                        'status_changed': datetime.now(tzutc()),
                    }).where(kernels.c.id == binding.kernel.kernel_id)
                    await kernel_db_conn.execute(query)

        await execute_with_retry(_finalize_scheduled)
        await self.registry.event_producer.produce_event(
            SessionScheduledEvent(sess_ctx.session_id, sess_ctx.session_creation_id)
        )

    async def prepare(
        self,
        context: None,
        source: AgentId,
        event: DoPrepareEvent,
    ) -> None:
        """
        Scan the scheduled sessions and perform the agent RPC calls to begin preparation of them.
        Each RPC calls are done in separate asyncio tasks.

        Session status transition: SCHEDULED -> PREPARING
        """
        known_slot_types = await self.shared_config.get_resource_slots()
        sched_ctx = SchedulingContext(
            self.registry,
            known_slot_types,
        )
        async with self.db.advisory_lock(AdvisoryLock.LOCKID_PREPARE):
            now = datetime.now(tzutc())

            async def _transition() -> Sequence[PendingSession]:
                async with self.db.begin() as conn:
                    update_query = (
                        sa.update(kernels)
                        .values({
                            'status': KernelStatus.PREPARING,
                            'status_changed': now,
                            'status_info': "",
                            'status_data': {},
                        })
                        .where(
                            (kernels.c.status == KernelStatus.SCHEDULED)
                        )
                        .returning(kernels.c.id)
                    )
                    rows = (await conn.execute(update_query)).fetchall()
                    if len(rows) == 0:
                        return []
                    target_kernel_ids = [r['id'] for r in rows]
                    select_query = (
                        PendingSession.base_query()
                        .where(
                            kernels.c.id.in_(target_kernel_ids)
                        )
                    )
                    rows = (await conn.execute(select_query)).fetchall()
                    return PendingSession.from_rows(rows)

            scheduled_sessions = await execute_with_retry(_transition)
            log.debug("prepare(): preparing {} session(s)", len(scheduled_sessions))
            async with aiotools.TaskGroup() as tg:
                for scheduled_session in scheduled_sessions:
                    await self.registry.event_producer.produce_event(
                        SessionPreparingEvent(
                            scheduled_session.session_id,
                            scheduled_session.session_creation_id,
                        )
                    )
                    tg.create_task(self.start_session(
                        sched_ctx,
                        scheduled_session,
                    ))

    async def start_session(
        self,
        sched_ctx: SchedulingContext,
        session: PendingSession,
    ) -> None:
        log_fmt = "prepare(s:{0.session_id}, type:{0.session_type}, name:{0.session_name}, " \
                  "ak:{0.access_key}, cluster_mode:{0.cluster_mode}): "
        log_args = (session, )
        log.debug(log_fmt + 'try-starting', *log_args)
        try:
            assert len(session.kernels) > 0
            await self.registry.start_session(sched_ctx, session)
        except Exception as e:
            status_data = convert_to_status_data(e, self.local_config['debug']['enabled'])
            log.warning(log_fmt + 'failed-starting: {1!r}', *log_args, status_data)
            # TODO: instead of instantly cancelling upon exception, we could mark it as
            #       SCHEDULED and retry within some limit using status_data.

            async def _update() -> None:
                async with self.db.begin() as db_conn:
                    for k in session.kernels:
                        await recalc_agent_resource_occupancy(db_conn, k.agent_id)
                    await _rollback_predicate_mutations(db_conn, sched_ctx, session)
                    now = datetime.now(tzutc())
                    query = kernels.update().values({
                        'status': KernelStatus.CANCELLED,
                        'status_changed': now,
                        'status_info': "failed-to-start",
                        'status_data': status_data,
                        'terminated_at': now,
                    }).where(kernels.c.session_id == session.session_id)
                    await db_conn.execute(query)

            await execute_with_retry(_update)
            await self.registry.event_producer.produce_event(
                SessionCancelledEvent(
                    session.session_id,
                    session.session_creation_id,
                    "failed-to-start",
                )
            )
            log.debug(log_fmt + 'cleanup-start-failure: begin', *log_args)
            try:
                async with self.db.begin_readonly() as db_conn:
                    query = (
                        sa.select([kernels.c.id, kernels.c.container_id])
                        .where(kernels.c.session_id == session.session_id)
                    )
                    rows = (await db_conn.execute(query)).fetchall()
                    cid_map = {row['id']: row['container_id'] for row in rows}
                destroyed_kernels = [
                    {
                        "agent": k.agent_id,
                        "agent_addr": k.agent_addr,
                        "id": k.kernel_id,
                        "container_id": cid_map[k.kernel_id],
                    }
                    for k in session.kernels
                ]
                await self.registry.destroy_session_lowlevel(
                    session.session_id, destroyed_kernels,
                )
            except Exception as destroy_err:
                log.error(log_fmt + 'cleanup-start-failure: error', *log_args, exc_info=destroy_err)
            finally:
                log.debug(log_fmt + 'cleanup-start-failure: done', *log_args)
        else:
            log.info(log_fmt + 'started', *log_args)


async def _list_pending_sessions(
    db_conn: SAConnection,
    sgroup_name: str,
) -> List[PendingSession]:
    query = (
        PendingSession.base_query()
        .where(
            (kernels.c.status == KernelStatus.PENDING) &
            (
                (kernels.c.scaling_group == sgroup_name) |
                (kernels.c.scaling_group.is_(None))
            )
        )
    )
    rows = (await db_conn.execute(query)).fetchall()
    return PendingSession.from_rows(rows)


async def _list_existing_sessions(
    db_conn: SAConnection,
    sgroup: str,
) -> List[ExistingSession]:
    query = (
        ExistingSession.base_query()
        .where(
            (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)) &
            (kernels.c.scaling_group == sgroup)
        )
    )
    rows = (await db_conn.execute(query)).fetchall()
    return ExistingSession.from_rows(rows)


async def _list_agents_by_sgroup(
    db_conn: SAConnection,
    sgroup_name: str,
) -> Sequence[AgentContext]:
    query = (
        sa.select([
            agents.c.id,
            agents.c.addr,
            agents.c.scaling_group,
            agents.c.available_slots,
            agents.c.occupied_slots,
        ])
        .select_from(agents)
        .where(
            (agents.c.status == AgentStatus.ALIVE) &
            (agents.c.scaling_group == sgroup_name) &
            (agents.c.schedulable == true())
        )
    )
    items = []
    for row in (await db_conn.execute(query)):
        item = AgentContext(
            row['id'],
            row['addr'],
            row['scaling_group'],
            row['available_slots'],
            row['occupied_slots'],
        )
        items.append(item)
    return items


async def _reserve_agent(
    sched_ctx: SchedulingContext,
    db_conn: SAConnection,
    scaling_group: str,
    agent_id: AgentId,
    requested_slots: ResourceSlot,
    extra_conds: Any = None,
) -> AgentAllocationContext:
    query = (
        sa.select([agents.c.occupied_slots])
        .select_from(agents)
        .where(agents.c.id == agent_id)
        .with_for_update()
    )
    if extra_conds is not None:
        query = query.where(extra_conds)
    current_occupied_slots = (await db_conn.execute(query)).scalar()
    if current_occupied_slots is None:
        raise RuntimeError(f"No agent matching condition: {extra_conds}")
    update_query = (
        sa.update(agents)
        .values({
            'occupied_slots': current_occupied_slots + requested_slots
        })
        .where(agents.c.id == agent_id)
    )
    await db_conn.execute(update_query)

    # Get the agent address for later RPC calls
    query = (sa.select([agents.c.addr])
               .where(agents.c.id == agent_id))
    agent_addr = await db_conn.scalar(query)
    assert agent_addr is not None

    return AgentAllocationContext(agent_id, agent_addr, scaling_group)


async def _rollback_predicate_mutations(
    db_conn: SAConnection,
    sched_ctx: SchedulingContext,
    session: PendingSession,
) -> None:
    """
    Rollback any changes performed by predicates.

    NOTE: We don't use the DB-level transaction rollback because we need to
    store the "ERROR" status to corresponding rows in the kernels table.
    """

    # Instead of decrementing concurrency_used, we recalculate the access_key's usage,
    # because asynchronous container launch failures and agent failures
    # (especially with multi-node multi-container cluster sessions)
    # may accumulate up multiple subtractions, resulting in
    # negative concurrency_occupied values.
    await recalc_concurrency_used(db_conn, session.access_key)



"""add_allowed_vfolder_hosts_to_domain_and_group

Revision ID: c401d78cc7b9
Revises: 3cf19d906e71
Create Date: 2019-06-26 11:34:55.426107

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'c401d78cc7b9'
down_revision = '3cf19d906e71'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('domains', sa.Column('allowed_vfolder_hosts',
                                       postgresql.ARRAY(sa.String()), nullable=True))
    op.add_column('groups', sa.Column('allowed_vfolder_hosts',
                                      postgresql.ARRAY(sa.String()), nullable=True))
    # ### end Alembic commands ###

    print('\nSet domain and group\'s allowed_vfolder_hosts with empty array.')
    connection = op.get_bind()
    query = ("UPDATE domains SET allowed_vfolder_hosts = '{}';")
    connection.execute(query)
    query = ("UPDATE groups SET allowed_vfolder_hosts = '{}';")
    connection.execute(query)

    op.alter_column('domains', column_name='allowed_vfolder_hosts', nullable=False)
    op.alter_column('groups', column_name='allowed_vfolder_hosts', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('groups', 'allowed_vfolder_hosts')
    op.drop_column('domains', 'allowed_vfolder_hosts')
    # ### end Alembic commands ###



"""add_creator_in_vfolders

Revision ID: 4b7b650bc30e
Revises: 202b6dcbc159
Create Date: 2019-10-30 00:23:57.085692

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4b7b650bc30e'
down_revision = '202b6dcbc159'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('vfolders', sa.Column('creator', sa.String(length=128), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('vfolders', 'creator')
    # ### end Alembic commands ###



"""Add mount_map column to kernel

Revision ID: d452bacd085c
Revises: 4b7b650bc30e
Create Date: 2019-11-19 14:43:12.728678

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql


# revision identifiers, used by Alembic.
revision = 'd452bacd085c'
down_revision = '4b7b650bc30e'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('kernels', sa.Column('mount_map', pgsql.JSONB(), nullable=True, default={}))


def downgrade():
    op.drop_column('kernels', 'mount_map')




"""
Resource preset APIs.
"""

import copy
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from decimal import Decimal
import functools
import json
import logging
import re
from typing import (
    Any,
    Iterable,
    TYPE_CHECKING,
    Tuple,
    MutableMapping,
)

import aiohttp
from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
from async_timeout import timeout as _timeout
from dateutil.tz import tzutc
import sqlalchemy as sa
import trafaret as t
import yarl

from ai.backend.common import validators as tx
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.utils import nmget
from ai.backend.common.types import DefaultForUnspecified, ResourceSlot

from ..models import (
    agents, resource_presets,
    domains, groups, kernels, users,
    AgentStatus,
    association_groups_users,
    query_allowed_sgroups,
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    RESOURCE_USAGE_KERNEL_STATUSES, LIVE_STATUS,
)
from .auth import auth_required, superadmin_required
from .exceptions import (
    InvalidAPIParameters,
)
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))

_json_loads = functools.partial(json.loads, parse_float=Decimal)


@auth_required
@atomic
async def list_presets(request: web.Request) -> web.Response:
    """
    Returns the list of all resource presets.
    """
    log.info('LIST_PRESETS (ak:{})', request['keypair']['access_key'])
    root_ctx: RootContext = request.app['_root.context']
    await root_ctx.shared_config.get_resource_slots()
    async with root_ctx.db.begin_readonly() as conn:
        query = (
            sa.select([resource_presets])
            .select_from(resource_presets)
        )
        # TODO: uncomment when we implement scaling group.
        # scaling_group = request.query.get('scaling_group')
        # if scaling_group is not None:
        #     query = query.where(resource_presets.c.scaling_group == scaling_group)
        resp: MutableMapping[str, Any] = {'presets': []}
        async for row in (await conn.stream(query)):
            preset_slots = row['resource_slots'].normalize_slots(ignore_unknown=True)
            resp['presets'].append({
                'name': row['name'],
                'shared_memory': str(row['shared_memory']) if row['shared_memory'] else None,
                'resource_slots': preset_slots.to_json(),
            })
        return web.json_response(resp, status=200)


@atomic
@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key('scaling_group', default=None): t.Null | t.String,
        t.Key('group', default='default'): t.String,
    }))
async def check_presets(request: web.Request, params: Any) -> web.Response:
    """
    Returns the list of all resource presets in the current scaling group,
    with additional information including allocatability of each preset,
    amount of total remaining resources, and the current keypair resource limits.
    """
    root_ctx: RootContext = request.app['_root.context']
    try:
        access_key = request['keypair']['access_key']
        resource_policy = request['keypair']['resource_policy']
        domain_name = request['user']['domain_name']
        # TODO: uncomment when we implement scaling group.
        # scaling_group = request.query.get('scaling_group')
        # assert scaling_group is not None, 'scaling_group parameter is missing.'
    except (json.decoder.JSONDecodeError, AssertionError) as e:
        raise InvalidAPIParameters(extra_msg=str(e.args[0]))
    known_slot_types = await root_ctx.shared_config.get_resource_slots()
    resp: MutableMapping[str, Any] = {
        'keypair_limits': None,
        'keypair_using': None,
        'keypair_remaining': None,
        'scaling_group_remaining': None,
        'scaling_groups': None,
        'presets': [],
    }
    log.info('CHECK_PRESETS (ak:{}, g:{}, sg:{})',
             request['keypair']['access_key'], params['group'], params['scaling_group'])

    async with root_ctx.db.begin_readonly() as conn:
        # Check keypair resource limit.
        keypair_limits = ResourceSlot.from_policy(resource_policy, known_slot_types)
        keypair_occupied = await root_ctx.registry.get_keypair_occupancy(access_key, conn=conn)
        keypair_remaining = keypair_limits - keypair_occupied

        # Check group resource limit and get group_id.
        j = sa.join(
            groups, association_groups_users,
            association_groups_users.c.group_id == groups.c.id,
        )
        query = (
            sa.select([groups.c.id, groups.c.total_resource_slots])
            .select_from(j)
            .where(
                (association_groups_users.c.user_id == request['user']['uuid']) &
                (groups.c.name == params['group']) &
                (domains.c.name == domain_name)
            )
        )
        result = await conn.execute(query)
        row = result.first()
        group_id = row['id']
        group_resource_slots = row['total_resource_slots']
        if group_id is None:
            raise InvalidAPIParameters('Unknown user group')
        group_resource_policy = {
            'total_resource_slots': group_resource_slots,
            'default_for_unspecified': DefaultForUnspecified.UNLIMITED
        }
        group_limits = ResourceSlot.from_policy(group_resource_policy, known_slot_types)
        group_occupied = await root_ctx.registry.get_group_occupancy(group_id, conn=conn)
        group_remaining = group_limits - group_occupied

        # Check domain resource limit.
        query = (sa.select([domains.c.total_resource_slots])
                   .where(domains.c.name == domain_name))
        domain_resource_slots = await conn.scalar(query)
        domain_resource_policy = {
            'total_resource_slots': domain_resource_slots,
            'default_for_unspecified': DefaultForUnspecified.UNLIMITED
        }
        domain_limits = ResourceSlot.from_policy(domain_resource_policy, known_slot_types)
        domain_occupied = await root_ctx.registry.get_domain_occupancy(domain_name, conn=conn)
        domain_remaining = domain_limits - domain_occupied

        # Take minimum remaining resources. There's no need to merge limits and occupied.
        # To keep legacy, we just merge all remaining slots into `keypair_remainig`.
        for slot in known_slot_types:
            keypair_remaining[slot] = min(
                keypair_remaining[slot],
                group_remaining[slot],
                domain_remaining[slot],
            )

        # Prepare per scaling group resource.
        sgroups = await query_allowed_sgroups(conn, domain_name, group_id, access_key)
        sgroup_names = [sg.name for sg in sgroups]
        if params['scaling_group'] is not None:
            if params['scaling_group'] not in sgroup_names:
                raise InvalidAPIParameters('Unknown scaling group')
            sgroup_names = [params['scaling_group']]
        per_sgroup = {
            sgname: {
                'using': ResourceSlot({k: Decimal(0) for k in known_slot_types.keys()}),
                'remaining': ResourceSlot({k: Decimal(0) for k in known_slot_types.keys()}),
            } for sgname in sgroup_names
        }

        # Per scaling group resource using from resource occupying kernels.
        query = (
            sa.select([kernels.c.occupied_slots, kernels.c.scaling_group])
            .select_from(kernels)
            .where(
                (kernels.c.user_uuid == request['user']['uuid']) &
                (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)) &
                (kernels.c.scaling_group.in_(sgroup_names))
            )
        )
        async for row in (await conn.stream(query)):
            per_sgroup[row['scaling_group']]['using'] += row['occupied_slots']

        # Per scaling group resource remaining from agents stats.
        sgroup_remaining = ResourceSlot({k: Decimal(0) for k in known_slot_types.keys()})
        query = (
            sa.select([agents.c.available_slots, agents.c.occupied_slots, agents.c.scaling_group])
            .select_from(agents)
            .where(
                (agents.c.status == AgentStatus.ALIVE) &
                (agents.c.scaling_group.in_(sgroup_names))
            )
        )
        agent_slots = []
        async for row in (await conn.stream(query)):
            remaining = row['available_slots'] - row['occupied_slots']
            remaining += ResourceSlot({k: Decimal(0) for k in known_slot_types.keys()})
            sgroup_remaining += remaining
            agent_slots.append(remaining)
            per_sgroup[row['scaling_group']]['remaining'] += remaining

        # Take maximum allocatable resources per sgroup.
        for sgname, sgfields in per_sgroup.items():
            for rtype, slots in sgfields.items():
                if rtype == 'remaining':
                    for slot in known_slot_types.keys():
                        if slot in slots:
                            slots[slot] = min(keypair_remaining[slot], slots[slot])
                per_sgroup[sgname][rtype] = slots.to_json()  # type: ignore  # it's serialization
        for slot in known_slot_types.keys():
            sgroup_remaining[slot] = min(keypair_remaining[slot], sgroup_remaining[slot])

        # Fetch all resource presets in the current scaling group.
        query = (
            sa.select([resource_presets])
            .select_from(resource_presets)
        )
        async for row in (await conn.stream(query)):
            # Check if there are any agent that can allocate each preset.
            allocatable = False
            preset_slots = row['resource_slots'].normalize_slots(ignore_unknown=True)
            for agent_slot in agent_slots:
                if agent_slot >= preset_slots and keypair_remaining >= preset_slots:
                    allocatable = True
                    break
            resp['presets'].append({
                'name': row['name'],
                'resource_slots': preset_slots.to_json(),
                'shared_memory': str(row['shared_memory']) if row['shared_memory'] is not None else None,
                'allocatable': allocatable,
            })

        # Return group resource status as NaN if not allowed.
        group_resource_visibility = \
            await root_ctx.shared_config.get_raw('config/api/resources/group_resource_visibility')
        group_resource_visibility = t.ToBool().check(group_resource_visibility)
        if not group_resource_visibility:
            group_limits = ResourceSlot({k: Decimal('NaN') for k in known_slot_types.keys()})
            group_occupied = ResourceSlot({k: Decimal('NaN') for k in known_slot_types.keys()})
            group_remaining = ResourceSlot({k: Decimal('NaN') for k in known_slot_types.keys()})

        resp['keypair_limits'] = keypair_limits.to_json()
        resp['keypair_using'] = keypair_occupied.to_json()
        resp['keypair_remaining'] = keypair_remaining.to_json()
        resp['group_limits'] = group_limits.to_json()
        resp['group_using'] = group_occupied.to_json()
        resp['group_remaining'] = group_remaining.to_json()
        resp['scaling_group_remaining'] = sgroup_remaining.to_json()
        resp['scaling_groups'] = per_sgroup
    return web.json_response(resp, status=200)


@server_status_required(READ_ALLOWED)
@superadmin_required
@atomic
async def recalculate_usage(request: web.Request) -> web.Response:
    """
    Update `keypairs.c.concurrency_used` and `agents.c.occupied_slots`.

    Those two values are sometimes out of sync. In that case, calling this API
    re-calculates the values for running containers and updates them in DB.
    """
    log.info('RECALCULATE_USAGE ()')
    root_ctx: RootContext = request.app['_root.context']
    await root_ctx.registry.recalc_resource_usage()
    return web.json_response({}, status=200)


async def get_container_stats_for_period(request: web.Request, start_date, end_date, group_ids=None):
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin_readonly(deferrable=True) as conn:
        j = (
            kernels
            .join(groups, groups.c.id == kernels.c.group_id)
            .join(users, users.c.uuid == kernels.c.user_uuid)
        )
        query = (
            sa.select([kernels, groups.c.name, users.c.email])
            .select_from(j)
            .where(
                # Filter sessions which existence period overlaps with requested period
                ((kernels.c.terminated_at >= start_date) & (kernels.c.created_at < end_date) &
                 (kernels.c.status.in_(RESOURCE_USAGE_KERNEL_STATUSES))) |
                # Or, filter running sessions which created before requested end_date
                ((kernels.c.created_at < end_date) & (kernels.c.status.in_(LIVE_STATUS)))
            )
            .order_by(sa.asc(kernels.c.terminated_at))
        )
        if group_ids:
            query = query.where(kernels.c.group_id.in_(group_ids))
        result = await conn.execute(query)
        rows = result.fetchall()

    objs_per_group = {}
    local_tz = root_ctx.shared_config['system']['timezone']

    for row in rows:
        group_id = str(row['group_id'])
        last_stat = row['last_stat']
        nfs = None
        if row['mounts'] is not None:
            nfs = list(set([mount[1] for mount in row['mounts']]))
        if row['terminated_at'] is None:
            used_time = used_days = None
        else:
            used_time = str(row['terminated_at'] - row['created_at'])
            used_days = (row['terminated_at'].astimezone(local_tz).toordinal() -
                         row['created_at'].astimezone(local_tz).toordinal() + 1)
        device_type = set()
        smp = 0
        gpu_mem_allocated = 0
        if row.attached_devices and row.attached_devices.get('cuda'):
            for dev_info in row.attached_devices['cuda']:
                if dev_info.get('model_name'):
                    device_type.add(dev_info['model_name'])
                smp += dev_info.get('smp', 0)
                gpu_mem_allocated += dev_info.get('mem', 0)
        gpu_allocated = 0
        if 'cuda.devices' in row.occupied_slots:
            gpu_allocated = row.occupied_slots['cuda.devices']
        if 'cuda.shares' in row.occupied_slots:
            gpu_allocated = row.occupied_slots['cuda.shares']
        c_info = {
            'id': str(row['id']),
            'container_id': row['container_id'],
            'domain_name': row['domain_name'],
            'group_id': str(row['group_id']),
            'group_name': row['name'],
            'name': row['session_name'],
            'access_key': row['access_key'],
            'email': row['email'],
            'agent': row['agent'],
            'cpu_allocated': float(row.occupied_slots.get('cpu', 0)),
            'cpu_used': float(nmget(last_stat, 'cpu_used.current', 0)),
            'mem_allocated': int(row.occupied_slots.get('mem', 0)),
            'mem_used': int(nmget(last_stat, 'mem.capacity', 0)),
            'shared_memory': int(nmget(row.resource_opts, 'shmem', 0)),
            'disk_allocated': 0,  # TODO: disk quota limit
            'disk_used': (int(nmget(last_stat, 'io_scratch_size/stats.max', 0, '/'))),
            'io_read': int(nmget(last_stat, 'io_read.current', 0)),
            'io_write': int(nmget(last_stat, 'io_write.current', 0)),
            'used_time': used_time,
            'used_days': used_days,
            'device_type': list(device_type),
            'smp': float(smp),
            'gpu_mem_allocated': float(gpu_mem_allocated),
            'gpu_allocated': float(gpu_allocated),  # devices or shares
            'nfs': nfs,
            'image_id': row['image'],  # TODO: image id
            'image_name': row['image'],
            'created_at': str(row['created_at']),
            'terminated_at': str(row['terminated_at']),
            'status': row['status'].name,
            'status_changed': str(row['status_changed']),
        }
        if group_id not in objs_per_group:
            objs_per_group[group_id] = {
                'domain_name': row['domain_name'],
                'g_id': group_id,
                'g_name': row['name'],  # this is group's name
                'g_cpu_allocated': c_info['cpu_allocated'],
                'g_cpu_used': c_info['cpu_used'],
                'g_mem_allocated': c_info['mem_allocated'],
                'g_mem_used': c_info['mem_used'],
                'g_shared_memory': c_info['shared_memory'],
                'g_disk_allocated': c_info['disk_allocated'],
                'g_disk_used': c_info['disk_used'],
                'g_io_read': c_info['io_read'],
                'g_io_write': c_info['io_write'],
                'g_device_type': copy.deepcopy(c_info['device_type']),
                'g_smp': c_info['smp'],
                'g_gpu_mem_allocated': c_info['gpu_mem_allocated'],
                'g_gpu_allocated': c_info['gpu_allocated'],
                'c_infos': [c_info],
            }
        else:
            objs_per_group[group_id]['g_cpu_allocated'] += c_info['cpu_allocated']
            objs_per_group[group_id]['g_cpu_used'] += c_info['cpu_used']
            objs_per_group[group_id]['g_mem_allocated'] += c_info['mem_allocated']
            objs_per_group[group_id]['g_mem_used'] += c_info['mem_used']
            objs_per_group[group_id]['g_shared_memory'] += c_info['shared_memory']
            objs_per_group[group_id]['g_disk_allocated'] += c_info['disk_allocated']
            objs_per_group[group_id]['g_disk_used'] += c_info['disk_used']
            objs_per_group[group_id]['g_io_read'] += c_info['io_read']
            objs_per_group[group_id]['g_io_write'] += c_info['io_write']
            for device in c_info['device_type']:
                if device not in objs_per_group[group_id]['g_device_type']:
                    g_dev_type = objs_per_group[group_id]['g_device_type']
                    g_dev_type.append(device)
                    objs_per_group[group_id]['g_device_type'] = list(set(g_dev_type))
            objs_per_group[group_id]['g_smp'] += c_info['smp']
            objs_per_group[group_id]['g_gpu_mem_allocated'] += c_info['gpu_mem_allocated']
            objs_per_group[group_id]['g_gpu_allocated'] += c_info['gpu_allocated']
            objs_per_group[group_id]['c_infos'].append(c_info)
    return list(objs_per_group.values())


@atomic
@server_status_required(READ_ALLOWED)
@superadmin_required
@check_api_params(
    t.Dict({
        tx.MultiKey('group_ids'): t.List(t.String) | t.Null,
        t.Key('month'): t.Regexp(r'^\d{6}', re.ASCII),
    }),
    loads=_json_loads)
async def usage_per_month(request: web.Request, params: Any) -> web.Response:
    """
    Return usage statistics of terminated containers for a specified month.
    The date/time comparison is done using the configured timezone.

    :param group_ids: If not None, query containers only in those groups.
    :param month: The year-month to query usage statistics. ex) "202006" to query for Jun 2020
    """
    log.info('USAGE_PER_MONTH (g:[{}], month:{})',
             ','.join(params['group_ids']), params['month'])
    root_ctx: RootContext = request.app['_root.context']
    local_tz = root_ctx.shared_config['system']['timezone']
    try:
        start_date = datetime.strptime(params['month'], '%Y%m').replace(tzinfo=local_tz)
        end_date = start_date + relativedelta(months=+1)
    except ValueError:
        raise InvalidAPIParameters(extra_msg='Invalid date values')
    resp = await get_container_stats_for_period(request, start_date, end_date, params['group_ids'])
    log.debug('container list are retrieved for month {0}', params['month'])
    return web.json_response(resp, status=200)


@atomic
@server_status_required(READ_ALLOWED)
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('group_id'): t.String | t.Null,
        t.Key('start_date'): t.Regexp(r'^\d{8}$', re.ASCII),
        t.Key('end_date'): t.Regexp(r'^\d{8}$', re.ASCII),
    }),
    loads=_json_loads)
async def usage_per_period(request: web.Request, params: Any) -> web.Response:
    """
    Return usage statistics of terminated containers belonged to the given group for a specified
    period in dates.
    The date/time comparison is done using the configured timezone.

    :param group_id: If not None, query containers only in the group.
    :param start_date str: "yyyymmdd" format.
    :param end_date str: "yyyymmdd" format.
    """
    root_ctx: RootContext = request.app['_root.context']
    group_id = params['group_id']
    local_tz = root_ctx.shared_config['system']['timezone']
    try:
        start_date = datetime.strptime(params['start_date'], '%Y%m%d').replace(tzinfo=local_tz)
        end_date = datetime.strptime(params['end_date'], '%Y%m%d').replace(tzinfo=local_tz)
        end_date = end_date + timedelta(days=1)  # include sessions in end_date
        if end_date - start_date > timedelta(days=100):
            raise InvalidAPIParameters('Cannot query more than 100 days')
    except ValueError:
        raise InvalidAPIParameters(extra_msg='Invalid date values')
    if end_date <= start_date:
        raise InvalidAPIParameters(extra_msg='end_date must be later than start_date.')
    log.info('USAGE_PER_MONTH (g:{}, start_date:{}, end_date:{})',
             group_id, start_date, end_date)
    group_ids = [group_id] if group_id is not None else None
    resp = await get_container_stats_for_period(request, start_date, end_date, group_ids=group_ids)
    log.debug('container list are retrieved from {0} to {1}', start_date, end_date)
    return web.json_response(resp, status=200)


async def get_time_binned_monthly_stats(request: web.Request, user_uuid=None):
    """
    Generate time-binned (15 min) stats for the last one month (2880 points).
    The structure of the result would be:

        [
          # [
          #     timestamp, num_sessions,
          #     cpu_allocated, mem_allocated, gpu_allocated,
          #     io_read, io_write, scratch_used,
          # ]
            [1562083808.657106, 1, 1.2, 1073741824, ...],
            [1562084708.657106, 2, 4.0, 1073741824, ...],
        ]

    Note that the timestamp is in UNIX-timestamp.
    """
    # Get all or user kernels for the last month from DB.
    time_window = 900  # 15 min
    now = datetime.now(tzutc())
    start_date = now - timedelta(days=30)
    root_ctx: RootContext = request.app['_root.context']
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([kernels])
            .select_from(kernels)
            .where(
                (kernels.c.terminated_at >= start_date) &
                (kernels.c.status.in_(RESOURCE_USAGE_KERNEL_STATUSES))
            )
            .order_by(sa.asc(kernels.c.created_at))
        )
        if user_uuid is not None:
            query = query.where(kernels.c.user_uuid == user_uuid)
        result = await conn.execute(query)
        rows = result.fetchall()

    # Build time-series of time-binned stats.
    rowcount = result.rowcount
    now_ts = now.timestamp()
    start_date_ts = start_date.timestamp()
    ts = start_date_ts
    idx = 0
    tseries = []
    # Iterate over each time window.
    while ts < now_ts:
        # Initialize the time-binned stats.
        num_sessions = 0
        cpu_allocated = 0
        mem_allocated = 0
        gpu_allocated = Decimal(0)
        io_read_bytes = 0
        io_write_bytes = 0
        disk_used = 0
        # Accumulate stats for containers overlapping with this time window.
        while idx < rowcount and \
              ts + time_window > rows[idx].created_at.timestamp() and \
              ts < rows[idx].terminated_at.timestamp():
            # Accumulate stats for overlapping containers in this time window.
            row = rows[idx]
            num_sessions += 1
            cpu_allocated += int(row.occupied_slots.get('cpu', 0))
            mem_allocated += int(row.occupied_slots.get('mem', 0))
            if 'cuda.devices' in row.occupied_slots:
                gpu_allocated += int(row.occupied_slots['cuda.devices'])
            if 'cuda.shares' in row.occupied_slots:
                gpu_allocated += Decimal(row.occupied_slots['cuda.shares'])
            if row.last_stat:
                io_read_bytes += int(nmget(row.last_stat, 'io_read.current', 0))
                io_write_bytes += int(nmget(row.last_stat, 'io_write.current', 0))
                disk_used += int(nmget(row.last_stat, 'io_scratch_size/stats.max', 0, '/'))
            idx += 1
        stat = {
            "date": ts,
            "num_sessions": {
                "value": num_sessions,
                "unit_hint": "count"
            },
            "cpu_allocated": {
                "value": cpu_allocated,
                "unit_hint": "count"
            },
            "mem_allocated": {
                "value": mem_allocated,
                "unit_hint": "bytes"
            },
            "gpu_allocated": {
                "value": float(gpu_allocated),
                "unit_hint": "count"
            },
            "io_read_bytes": {
                "value": io_read_bytes,
                "unit_hint": "bytes"
            },
            "io_write_bytes": {
                "value": io_write_bytes,
                "unit_hint": "bytes"
            },
            "disk_used": {
                "value ": disk_used,
                "unit_hint": "bytes"
            }
        }
        tseries.append(stat)
        ts += time_window
    return tseries


@server_status_required(READ_ALLOWED)
@auth_required
async def user_month_stats(request: web.Request) -> web.Response:
    """
    Return time-binned (15 min) stats for terminated user sessions
    over last 30 days.
    """
    access_key = request['keypair']['access_key']
    user_uuid = request['user']['uuid']
    log.info('USER_LAST_MONTH_STATS (ak:{}, u:{})', access_key, user_uuid)
    stats = await get_time_binned_monthly_stats(request, user_uuid=user_uuid)
    return web.json_response(stats, status=200)


@server_status_required(READ_ALLOWED)
@superadmin_required
async def admin_month_stats(request: web.Request) -> web.Response:
    """
    Return time-binned (15 min) stats for all terminated sessions
    over last 30 days.
    """
    log.info('ADMIN_LAST_MONTH_STATS ()')
    stats = await get_time_binned_monthly_stats(request, user_uuid=None)
    return web.json_response(stats, status=200)


async def get_watcher_info(request: web.Request, agent_id: str) -> dict:
    """
    Get watcher information.

    :return addr: address of agent watcher (eg: http://127.0.0.1:6009)
    :return token: agent watcher token ("insecure" if not set in config server)
    """
    root_ctx: RootContext = request.app['_root.context']
    token = root_ctx.shared_config['watcher']['token']
    if token is None:
        token = 'insecure'
    agent_ip = await root_ctx.shared_config.etcd.get(f'nodes/agents/{agent_id}/ip')
    watcher_port = await root_ctx.shared_config.etcd.get(
        f'nodes/agents/{agent_id}/watcher_port'
    )
    if watcher_port is None:
        watcher_port = 6009
    # TODO: watcher scheme is assumed to be http
    addr = yarl.URL(f'http://{agent_ip}:{watcher_port}')
    return {
        'addr': addr,
        'token': token,
    }


@server_status_required(READ_ALLOWED)
@superadmin_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(['agent_id', 'agent']): t.String,
    }))
async def get_watcher_status(request: web.Request, params: Any) -> web.Response:
    log.info('GET_WATCHER_STATUS ()')
    watcher_info = await get_watcher_info(request, params['agent_id'])
    connector = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=connector) as sess:
        with _timeout(5.0):
            headers = {'X-BackendAI-Watcher-Token': watcher_info['token']}
            async with sess.get(watcher_info['addr'], headers=headers) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    return web.json_response(data, status=resp.status)
                else:
                    data = await resp.text()
                    return web.Response(text=data, status=resp.status)


@server_status_required(READ_ALLOWED)
@superadmin_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(['agent_id', 'agent']): t.String,
    }))
async def watcher_agent_start(request: web.Request, params: Any) -> web.Response:
    log.info('WATCHER_AGENT_START ()')
    watcher_info = await get_watcher_info(request, params['agent_id'])
    connector = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=connector) as sess:
        with _timeout(20.0):
            watcher_url = watcher_info['addr'] / 'agent/start'
            headers = {'X-BackendAI-Watcher-Token': watcher_info['token']}
            async with sess.post(watcher_url, headers=headers) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    return web.json_response(data, status=resp.status)
                else:
                    data = await resp.text()
                    return web.Response(text=data, status=resp.status)


@server_status_required(READ_ALLOWED)
@superadmin_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(['agent_id', 'agent']): t.String,
    }))
async def watcher_agent_stop(request: web.Request, params: Any) -> web.Response:
    log.info('WATCHER_AGENT_STOP ()')
    watcher_info = await get_watcher_info(request, params['agent_id'])
    connector = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=connector) as sess:
        with _timeout(20.0):
            watcher_url = watcher_info['addr'] / 'agent/stop'
            headers = {'X-BackendAI-Watcher-Token': watcher_info['token']}
            async with sess.post(watcher_url, headers=headers) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    return web.json_response(data, status=resp.status)
                else:
                    data = await resp.text()
                    return web.Response(text=data, status=resp.status)


@server_status_required(READ_ALLOWED)
@superadmin_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(['agent_id', 'agent']): t.String,
    }))
async def watcher_agent_restart(request: web.Request, params: Any) -> web.Response:
    log.info('WATCHER_AGENT_RESTART ()')
    watcher_info = await get_watcher_info(request, params['agent_id'])
    connector = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=connector) as sess:
        with _timeout(20.0):
            watcher_url = watcher_info['addr'] / 'agent/restart'
            headers = {'X-BackendAI-Watcher-Token': watcher_info['token']}
            async with sess.post(watcher_url, headers=headers) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    return web.json_response(data, status=resp.status)
                else:
                    data = await resp.text()
                    return web.Response(text=data, status=resp.status)


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app['api_versions'] = (4,)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    cors.add(add_route('GET',  '/presets', list_presets))
    cors.add(add_route('POST', '/check-presets', check_presets))
    cors.add(add_route('POST', '/recalculate-usage', recalculate_usage))
    cors.add(add_route('GET',  '/usage/month', usage_per_month))
    cors.add(add_route('GET',  '/usage/period', usage_per_period))
    cors.add(add_route('GET',  '/stats/user/month', user_month_stats))
    cors.add(add_route('GET',  '/stats/admin/month', admin_month_stats))
    cors.add(add_route('GET',  '/watcher', get_watcher_status))
    cors.add(add_route('POST', '/watcher/agent/start', watcher_agent_start))
    cors.add(add_route('POST', '/watcher/agent/stop', watcher_agent_stop))
    cors.add(add_route('POST', '/watcher/agent/restart', watcher_agent_restart))
    return app, []



from __future__ import annotations

import logging
import sys
import traceback
from typing import Any, Mapping, TYPE_CHECKING

from ai.backend.common.events import AgentErrorEvent
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    AgentId,
    LogSeverity,
)
from ai.backend.common.plugin.monitor import AbstractErrorReporterPlugin

from ..models import error_logs

if TYPE_CHECKING:
    from ai.backend.manager.api.context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))


class ErrorMonitor(AbstractErrorReporterPlugin):

    async def init(self, context: Any = None) -> None:
        root_ctx: RootContext = context['_root.context']  # type: ignore
        self.event_dispatcher = root_ctx.event_dispatcher
        self._evh = self.event_dispatcher.consume(AgentErrorEvent, None, self.handle_agent_error)
        self.db = root_ctx.db

    async def cleanup(self) -> None:
        self.event_dispatcher.unconsume(self._evh)

    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        pass

    async def capture_message(self, message: str) -> None:
        pass

    async def capture_exception(
        self,
        exc_instance: Exception = None,
        context: Mapping[str, Any] = None,
    ) -> None:
        if exc_instance:
            tb = exc_instance.__traceback__
        else:
            _, sys_exc_instance, tb = sys.exc_info()
            if (
                isinstance(sys_exc_instance, BaseException)
                and not isinstance(sys_exc_instance, Exception)
            ):
                # bypass BaseException as they are used for controlling the process/coroutine lifecycles
                # instead of indicating actual errors
                return
            exc_instance = sys_exc_instance
        exc_type: Any = type(exc_instance)

        if context is None or 'severity' not in context:
            severity = LogSeverity.ERROR
        else:
            severity = context['severity']
        if context is None or 'user' not in context:
            user = None
        else:
            user = context['user']

        async with self.db.begin() as conn:
            query = error_logs.insert().values({
                'severity': severity,
                'source': 'manager',
                'user': user,
                'message': ''.join(traceback.format_exception_only(exc_type, exc_instance)).strip(),
                'context_lang': 'python',
                'context_env': context,
                'traceback': ''.join(traceback.format_tb(tb)).strip()
            })
            await conn.execute(query)
        log.debug('Manager log collected: {}', str(exc_instance))

    async def handle_agent_error(
        self,
        context: None,
        source: AgentId,
        event: AgentErrorEvent,
    ) -> None:
        async with self.db.begin() as conn:
            query = error_logs.insert().values({
                'severity': event.severity,
                'source': source,
                'user': event.user,
                'message': event.message,
                'context_lang': 'python',
                'context_env': event.context_env,
                'traceback': event.traceback
            })
            await conn.execute(query)
        log.debug(
            'collected AgentErrorEvent: [{}:{}] {}',
            source, event.severity, event.message,
        )



"""add-batch-session

Revision ID: c092dabf3ee5
Revises: c1409ad0e8da
Create Date: 2019-08-01 15:18:20.306290

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'c092dabf3ee5'
down_revision = '48ab2dfefba9'
branch_labels = None
depends_on = None


sessiontypes = postgresql.ENUM(
    'INTERACTIVE', 'BATCH',
    name='sessiontypes')


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('kernels', sa.Column('startup_command', sa.Text(), nullable=True))
    op.drop_column('kernels', 'type')
    op.add_column('kernels',
                  sa.Column('sess_type', sa.Enum('INTERACTIVE', 'BATCH', name='sessiontypes'),
                            nullable=False, server_default='INTERACTIVE'))
    op.create_index(op.f('ix_kernels_sess_type'), 'kernels', ['sess_type'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_kernels_sess_type'), table_name='kernels')
    op.drop_column('kernels', 'sess_type')
    op.add_column('kernels',
                  sa.Column('type', sa.Enum('INTERACTIVE', 'BATCH', name='sessiontypes'),
                            nullable=False, server_default='INTERACTIVE'))
    op.drop_column('kernels', 'startup_command')
    # ### end Alembic commands ###



"""Add clusterized column to agents table

Revision ID: 1e673659b283
Revises: d5cc54fd36b5
Create Date: 2020-01-07 17:52:51.771357

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1e673659b283'
down_revision = 'd5cc54fd36b5'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'agents',
        sa.Column('clusterized', sa.Boolean, default=False)
    )


def downgrade():
    op.drop_column('agents', 'clusterized')



"""allow_kernels_scaling_group_nullable

Revision ID: 9bd986a75a2a
Revises: 513164749de4
Create Date: 2019-09-20 14:39:57.761791

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9bd986a75a2a'
down_revision = '513164749de4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('kernels', 'scaling_group',
               existing_type=sa.VARCHAR(length=64),
               nullable=True,
               existing_server_default=sa.text("'default'::character varying"))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('kernels', 'scaling_group',
               existing_type=sa.VARCHAR(length=64),
               nullable=False,
               existing_server_default=sa.text("'default'::character varying"))
    # ### end Alembic commands ###



"""add state column to vfolder_invitations

Revision ID: f9971fbb34d9
Revises: 185852ff9872
Create Date: 2018-07-12 23:30:14.942845

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f9971fbb34d9'
down_revision = '185852ff9872'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('vfolder_invitations', sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=True))
    op.add_column('vfolder_invitations', sa.Column('state', sa.String(length=10), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('vfolder_invitations', 'state')
    op.drop_column('vfolder_invitations', 'created_at')
    # ### end Alembic commands ###



"""add_clone_allowed_to_vfolders

Revision ID: d463fc5d6109
Revises: 0d553d59f369
Create Date: 2020-09-01 16:57:55.339619

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd463fc5d6109'
down_revision = '0d553d59f369'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('vfolders', sa.Column('clone_allowed', sa.Boolean(), nullable=True))
    op.execute("UPDATE vfolders SET clone_allowed = false")
    op.alter_column('vfolders', 'clone_allowed', nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('vfolders', 'clone_allowed')
    # ### end Alembic commands ###



from __future__ import annotations

from abc import ABCMeta, abstractmethod
import logging
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Protocol,
    Sequence,
    Set,
)
import uuid

import attr
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.sql import Select, ColumnElement
from sqlalchemy.engine.row import Row
import trafaret as t

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.docker import (
    ImageRef,
)
from ai.backend.common.types import (
    AccessKey,
    AgentId,
    ClusterMode,
    KernelId,
    SessionId,
    SessionTypes,
    ResourceSlot,
    SlotName,
    SlotTypes,
)

from ..defs import DEFAULT_ROLE
from ..models import (
    kernels, keypairs,
)
from ..registry import AgentRegistry

log = BraceStyleAdapter(logging.getLogger('ai.backend.manager.scheduler'))


def merge_resource(
    target: MutableMapping[str, Any],
    update: MutableMapping[str, Any],
) -> None:
    for k in update.keys():
        if k in target.keys():
            target[k] += update[k]
        else:
            target[k] = update[k]


@attr.s(auto_attribs=True, slots=True)
class AgentAllocationContext:
    agent_id: AgentId
    agent_addr: str
    scaling_group: str


@attr.s(auto_attribs=True, slots=True)
class AgentContext:
    agent_id: AgentId
    agent_addr: str
    scaling_group: str
    available_slots: ResourceSlot
    occupied_slots: ResourceSlot


@attr.s(auto_attribs=True, slots=True)
class ScheduleDecision:
    agent_id: AgentId
    kernel_id: KernelId


@attr.s(auto_attribs=True, slots=True)
class SchedulingContext:
    """
    Context for each scheduling decision.
    """
    registry: AgentRegistry
    known_slot_types: Mapping[SlotName, SlotTypes]


@attr.s(auto_attribs=True, slots=True)
class ExistingSession:
    kernels: List[KernelInfo]
    access_key: AccessKey
    session_id: uuid.UUID
    session_type: SessionTypes
    session_name: str
    cluster_mode: ClusterMode
    cluster_size: int
    domain_name: str
    group_id: uuid.UUID
    scaling_group: str
    occupying_slots: ResourceSlot

    @classmethod
    def db_cols(cls) -> Set[ColumnElement]:
        return {
            kernels.c.id,
            kernels.c.status,
            kernels.c.access_key,
            kernels.c.session_id,
            kernels.c.session_type,
            kernels.c.session_name,
            kernels.c.cluster_mode,
            kernels.c.cluster_size,
            kernels.c.cluster_role,
            kernels.c.domain_name,
            kernels.c.group_id,
            kernels.c.scaling_group,
            kernels.c.occupied_slots,
        }

    @classmethod
    def base_query(cls) -> Select:
        return (
            sa.select(
                list(cls.db_cols() | KernelInfo.db_cols())
            )
            .select_from(kernels)
            .order_by(kernels.c.created_at)
        )

    @classmethod
    def from_row(cls, row: Row) -> ExistingSession:
        return ExistingSession(
            kernels=[],
            access_key=row['access_key'],
            session_id=row['session_id'],
            session_type=row['session_type'],
            session_name=row['session_name'],
            cluster_mode=row['cluster_mode'],
            cluster_size=row['cluster_size'],
            domain_name=row['domain_name'],
            group_id=row['group_id'],
            scaling_group=row['scaling_group'],
            occupying_slots=ResourceSlot(),
        )

    @classmethod
    def from_rows(cls, rows: Sequence[Row]) -> List[ExistingSession]:
        items: Dict[str, ExistingSession] = {}
        for row in rows:
            if row['cluster_role'] == "main":
                items[row['session_id']] = cls.from_row(row)
        for row in rows:
            session_id = row['session_id']
            if session_id not in items:
                # In some cases, sub containers are still RUNNING
                # even though main container is TERMINATED.
                # To circumvent this edge case, we skip if main container
                # is not registered in `items`.
                continue
            session = items[session_id]
            session.kernels.append(KernelInfo.from_row(row))
            session.occupying_slots += row['occupied_slots']  # type: ignore
        return list(items.values())


@attr.s(auto_attribs=True, slots=True)
class PendingSession:
    """
    Context for individual session-related information used during scheduling.
    Resource parameters defined here should contain total amount of resources
    for all kernels in one session.
    """
    kernels: List[KernelInfo]
    access_key: AccessKey
    session_id: SessionId
    session_creation_id: str
    session_type: SessionTypes
    session_name: str
    cluster_mode: ClusterMode
    cluster_size: int
    domain_name: str
    group_id: uuid.UUID
    status_data: Mapping[str, Any]
    scaling_group: str
    resource_policy: str
    resource_opts: Mapping[str, Any]
    requested_slots: ResourceSlot
    target_sgroup_names: MutableSequence[str]
    environ: MutableMapping[str, str]
    mounts: Sequence[str]
    mount_map: Mapping[str, str]
    bootstrap_script: Optional[str]
    startup_command: Optional[str]
    internal_data: Optional[MutableMapping[str, Any]]
    preopen_ports: List[int]

    @property
    def main_kernel_id(self) -> KernelId:
        for k in self.kernels:
            if k.cluster_role == DEFAULT_ROLE:
                return k.kernel_id
        raise RuntimeError('Unable to get the main kernel ID')

    @classmethod
    def db_cols(cls) -> Set[ColumnElement]:
        return {
            kernels.c.id,
            kernels.c.access_key,
            kernels.c.session_creation_id,
            kernels.c.session_id,
            kernels.c.session_type,
            kernels.c.session_name,
            kernels.c.cluster_mode,
            kernels.c.cluster_size,
            kernels.c.domain_name,
            kernels.c.group_id,
            kernels.c.status_data,
            kernels.c.scaling_group,
            keypairs.c.resource_policy,
            kernels.c.occupied_slots,
            kernels.c.internal_data,
            kernels.c.resource_opts,
            kernels.c.environ,
            kernels.c.mounts,
            kernels.c.mount_map,
            kernels.c.bootstrap_script,
            kernels.c.startup_command,
            kernels.c.preopen_ports,
        }

    @classmethod
    def base_query(cls) -> Select:
        return (
            sa.select(
                list(cls.db_cols() | KernelInfo.db_cols())
            )
            .select_from(sa.join(
                kernels, keypairs,
                keypairs.c.access_key == kernels.c.access_key
            ))
            .order_by(kernels.c.created_at)
        )

    @classmethod
    def from_row(cls, row: Row) -> PendingSession:
        return cls(
            kernels=[],
            access_key=row['access_key'],
            session_creation_id=row['session_creation_id'],
            session_id=row['session_id'],
            session_type=row['session_type'],
            session_name=row['session_name'],
            cluster_mode=row['cluster_mode'],
            cluster_size=row['cluster_size'],
            domain_name=row['domain_name'],
            group_id=row['group_id'],
            status_data=row['status_data'],
            scaling_group=row['scaling_group'],
            resource_policy=row['resource_policy'],
            resource_opts={},
            requested_slots=ResourceSlot(),
            internal_data=row['internal_data'],
            target_sgroup_names=[],
            environ={
                k: v for k, v
                in map(lambda s: s.split('=', maxsplit=1), row['environ'])
            },
            mounts=row['mounts'],
            mount_map=row['mount_map'],
            bootstrap_script=row['bootstrap_script'],
            startup_command=row['startup_command'],
            preopen_ports=row['preopen_ports'],
        )

    @classmethod
    def from_rows(cls, rows: Sequence[Row]) -> List[PendingSession]:
        items: Dict[SessionId, PendingSession] = {}
        for row in rows:
            if row['cluster_role'] == "main":
                items[row['session_id']] = cls.from_row(row)
        for row in rows:
            session = items[row['session_id']]
            session.kernels.append(KernelInfo.from_row(row))
            session.requested_slots += row['occupied_slots']  # type: ignore
            merge_resource(session.resource_opts, row['resource_opts'])  # type: ignore
        return list(items.values())


@attr.s(auto_attribs=True, slots=True)
class KernelInfo:
    """
    Representing invididual kernel info.
    Resource parameters defined here should contain single value of resource
    for each kernel.
    """
    kernel_id: KernelId
    session_id: SessionId
    access_key: AccessKey
    agent_id: AgentId
    agent_addr: str
    cluster_role: str
    cluster_idx: int
    cluster_hostname: str
    image_ref: ImageRef
    resource_opts: Mapping[str, Any]
    requested_slots: ResourceSlot
    bootstrap_script: Optional[str]
    startup_command: Optional[str]

    def __str__(self):
        return f'{self.kernel_id}#{self.cluster_role}{self.cluster_idx}'

    @classmethod
    def db_cols(cls) -> Set[ColumnElement]:
        return {
            kernels.c.id,
            kernels.c.session_id,
            kernels.c.access_key,
            kernels.c.agent,       # for scheduled kernels
            kernels.c.agent_addr,  # for scheduled kernels
            kernels.c.cluster_role,
            kernels.c.cluster_idx,
            kernels.c.cluster_hostname,
            kernels.c.image,
            kernels.c.registry,
            kernels.c.resource_opts,
            kernels.c.occupied_slots,
            kernels.c.bootstrap_script,
            kernels.c.startup_command,
        }

    @classmethod
    def from_row(cls, row: Row) -> KernelInfo:
        return cls(
            kernel_id=row['id'],
            session_id=row['session_id'],
            access_key=row['access_key'],
            agent_id=row['agent'],
            agent_addr=row['agent_addr'],
            cluster_role=row['cluster_role'],
            cluster_idx=row['cluster_idx'],
            cluster_hostname=row['cluster_hostname'],
            image_ref=ImageRef(row['image'], [row['registry']]),
            resource_opts=row['resource_opts'],
            requested_slots=row['occupied_slots'],
            bootstrap_script=row['bootstrap_script'],
            startup_command=row['startup_command'],
        )


@attr.s(auto_attribs=True, slots=True)
class KernelAgentBinding:
    kernel: KernelInfo
    agent_alloc_ctx: AgentAllocationContext


@attr.s(auto_attribs=True, slots=True)
class PredicateResult:
    passed: bool
    message: Optional[str] = None
    permanent: bool = False


class SchedulingPredicate(Protocol):
    async def __call__(
        self,
        db_conn: SAConnection,
        sched_ctx: SchedulingContext,
        sess_ctx: PendingSession,
    ) -> PredicateResult:
        ...


class AbstractScheduler(metaclass=ABCMeta):

    """
    Interface for scheduling algorithms where the
    ``schedule()`` method is a pure function.
    """

    config: Mapping[str, Any]
    config_iv: t.Dict

    def __init__(self, config: Mapping[str, Any]) -> None:
        self.config = self.config_iv.check(config)

    @abstractmethod
    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[PendingSession],
        existing_sessions: Sequence[ExistingSession],
    ) -> Optional[SessionId]:
        """
        Pick a session to try schedule.
        This is where the queueing semantics is implemented such as prioritization.
        """
        return None

    @abstractmethod
    def assign_agent_for_session(
        self,
        possible_agents: Sequence[AgentContext],
        pending_session: PendingSession,
    ) -> Optional[AgentId]:
        """
        Assign an agent for the entire session, only considering the total requested
        slots of the session.  This is used for both single-container sessions and
        single-node multi-container sessions.

        In single-node multi-container sessions, all sub-containers are spawned by
        slicing the assigned agent's resource.
        """
        return None

    @abstractmethod
    def assign_agent_for_kernel(
        self,
        possible_agents: Sequence[AgentContext],
        pending_kernel: KernelInfo,
    ) -> Optional[AgentId]:
        """
        Assign an agent for a kernel of the session.
        This may be called multiple times for multi-node multi-container sessions.
        """
        return None



"""dynamic-resource-slots

Revision ID: f0f4ee907155
Revises: ff4bfca66bf8
Create Date: 2019-01-27 17:05:13.997279

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'f0f4ee907155'
down_revision = 'ff4bfca66bf8'
branch_labels = None
depends_on = None


def upgrade():

    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    op.alter_column('kernels', 'service_ports',
                    existing_type=sa.JSON(),
                    type_=postgresql.JSONB(),
                    postgresql_using='CAST(service_ports AS jsonb)')
    op.add_column('agents', sa.Column('available_slots', postgresql.JSONB(),
                                      nullable=False,
                                      server_default=sa.text("'{}'::jsonb")))
    op.add_column('agents', sa.Column('occupied_slots', postgresql.JSONB(),
                                      nullable=False,
                                      server_default=sa.text("'{}'::jsonb")))
    query = '''
    UPDATE agents SET available_slots = json_strip_nulls(json_build_object(
        'cpu', cpu_slots,
        'mem', mem_slots::text || 'g'
    ));
    UPDATE agents SET available_slots = available_slots || json_build_object(
        'cuda.device', gpu_slots
    )::jsonb
    WHERE gpu_slots > 0;
    UPDATE agents SET available_slots = available_slots || json_build_object(
        'tpu.device', tpu_slots
    )::jsonb
    WHERE tpu_slots > 0;

    UPDATE agents SET occupied_slots = json_strip_nulls(json_build_object(
        'cpu', used_cpu_slots,
        'mem', used_mem_slots::text || 'g'
    ));
    UPDATE agents SET occupied_slots = occupied_slots || json_build_object(
        'cuda.device', used_gpu_slots
    )::jsonb
    WHERE used_gpu_slots > 0;
    UPDATE agents SET occupied_slots = occupied_slots || json_build_object(
        'tpu.device', used_tpu_slots
    )::jsonb
    WHERE used_tpu_slots > 0;
    '''
    connection.execute(query)
    op.drop_column('agents', 'cpu_slots')
    op.drop_column('agents', 'mem_slots')
    op.drop_column('agents', 'gpu_slots')
    op.drop_column('agents', 'tpu_slots')
    op.drop_column('agents', 'used_cpu_slots')
    op.drop_column('agents', 'used_mem_slots')
    op.drop_column('agents', 'used_gpu_slots')
    op.drop_column('agents', 'used_tpu_slots')

    op.add_column('kernels', sa.Column('occupied_slots', postgresql.JSONB(),
                                       nullable=False,
                                       server_default=sa.text("'{}'::jsonb")))
    op.add_column('kernels', sa.Column('occupied_shares', postgresql.JSONB(),
                                       nullable=False,
                                       server_default=sa.text("'{}'::jsonb")))
    query = '''
    UPDATE kernels SET occupied_slots = json_build_object(
        'cpu', cpu_slot,
        'mem', mem_slot,
        'cuda.device', gpu_slot,
        'tpu.device', tpu_slot
    );
    UPDATE kernels SET occupied_shares = json_build_object(
        'cpu', cpu_set,
        'mem', mem_slot,
        'cuda.device', '{}'::json,
        'tpu.device', '{}'::json
    );
    '''
    connection.execute(query)
    op.drop_column('kernels', 'cpu_slot')
    op.drop_column('kernels', 'mem_slot')
    op.drop_column('kernels', 'gpu_slot')
    op.drop_column('kernels', 'tpu_slot')
    op.drop_column('kernels', 'cpu_set')
    op.drop_column('kernels', 'gpu_set')
    op.drop_column('kernels', 'tpu_set')
    # ### end Alembic commands ###


def downgrade():
    op.alter_column('kernels', 'service_ports',
                    existing_type=postgresql.JSONB(),
                    type_=sa.JSON(),
                    postgresql_using='CAST(service_ports AS json)')
    connection = op.get_bind()

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('kernels', sa.Column('cpu_set', postgresql.ARRAY(sa.INTEGER()),
                                       autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('gpu_set', postgresql.ARRAY(sa.INTEGER()),
                                       autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('tpu_set', postgresql.ARRAY(sa.INTEGER()),
                                       autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('cpu_slot',
                                       postgresql.DOUBLE_PRECISION(precision=53),
                                       autoincrement=False, nullable=False,
                                       server_default='0'))
    op.add_column('kernels', sa.Column('mem_slot', sa.BIGINT(),
                                       autoincrement=False, nullable=False,
                                       server_default='0'))
    op.add_column('kernels', sa.Column('gpu_slot',
                                       postgresql.DOUBLE_PRECISION(precision=53),
                                       autoincrement=False, nullable=False,
                                       server_default='0'))
    op.add_column('kernels', sa.Column('tpu_slot',
                                       postgresql.DOUBLE_PRECISION(precision=53),
                                       autoincrement=False, nullable=False,
                                       server_default='0'))
    query = '''
    UPDATE kernels
    SET
        cpu_set = (
            SELECT coalesce(array_agg(v::text::int), '{}')
            FROM json_array_elements((occupied_shares->>'cpu')::json) v),
        gpu_set = (
            SELECT coalesce(array_agg(v::text::int), '{}')
            FROM json_array_elements((occupied_shares->>'cuda.device')::json) v),
        tpu_set = (
            SELECT coalesce(array_agg(v::text::int), '{}')
            FROM json_array_elements((occupied_shares->>'tpu.device')::json) v)
    ;
    '''
    connection.execute(query)
    query = '''
    UPDATE kernels
    SET
        cpu_slot = coalesce((occupied_slots->>'cpu')::text::float, 0),
        mem_slot = coalesce((occupied_slots->>'mem')::text::bigint, 0),
        gpu_slot = coalesce((occupied_slots->>'cuda.device')::text::float, 0),
        tpu_slot = coalesce((occupied_slots->>'tpu.device')::text::float, 0)
    ;
    '''
    connection.execute(query)
    op.drop_column('kernels', 'occupied_shares')
    op.drop_column('kernels', 'occupied_slots')

    op.add_column('agents', sa.Column('gpu_slots',
                                      postgresql.DOUBLE_PRECISION(precision=53),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('used_cpu_slots',
                                      postgresql.DOUBLE_PRECISION(precision=53),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('tpu_slots',
                                      postgresql.DOUBLE_PRECISION(precision=53),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('used_mem_slots', sa.BIGINT(),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('cpu_slots',
                                      postgresql.DOUBLE_PRECISION(precision=53),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('used_gpu_slots',
                                      postgresql.DOUBLE_PRECISION(precision=53),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('mem_slots', sa.BIGINT(),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    op.add_column('agents', sa.Column('used_tpu_slots',
                                      postgresql.DOUBLE_PRECISION(precision=53),
                                      autoincrement=False, nullable=False,
                                      server_default='0'))
    query = '''
    UPDATE agents
    SET
        cpu_slots = coalesce((available_slots->>'cpu')::text::float, 0),
        mem_slots = coalesce((available_slots->>'mem')::text::bigint, 0),
        gpu_slots = coalesce((available_slots->>'cuda.device')::text::float, 0),
        tpu_slots = coalesce((available_slots->>'tpu.device')::text::float, 0)
    ;
    '''
    connection.execute(query)
    query = '''
    UPDATE agents
    SET
        used_cpu_slots = coalesce((occupied_slots->>'cpu')::text::float, 0),
        used_mem_slots = coalesce((occupied_slots->>'mem')::text::bigint, 0),
        used_gpu_slots = coalesce((occupied_slots->>'cuda.device')::text::float, 0),
        used_tpu_slots = coalesce((occupied_slots->>'tpu.device')::text::float, 0)
    ;
    '''
    connection.execute(query)
    op.drop_column('agents', 'occupied_slots')
    op.drop_column('agents', 'available_slots')
    # ### end Alembic commands ###
